# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Inputs produce a list of Items that are used to make changes to the codebase.
These Items can represent anything and only need to be understood by the components of the Schema.
Most use cases will return FileItems.
"""



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the SchemaNameCondition."""

from __future__ import annotations

from typing import ClassVar, List

from autotransform.change.base import Change
from autotransform.step.condition.base import ComparisonCondition, ConditionName
from autotransform.step.condition.comparison import ComparisonType


class SchemaNameCondition(ComparisonCondition[str]):
    """A condition which checks the name of the Schema that produced a change against the supplied
    name, using the supplied comparison.

    Attributes:
        comparison (ComparisonType): The type of comparison to perform.
        value (str | List[str]): The schema name(s) to compare against.
        name (ClassVar[ConditionName]): The name of the Component.
    """

    comparison: ComparisonType
    value: str | List[str]

    name: ClassVar[ConditionName] = ConditionName.SCHEMA_NAME

    def get_val_from_change(self, change: Change) -> str:
        """Gets the Schema name from the Change.

        Args:
            change (Change): The Change the Condition is checking.

        Returns:
            str: The name of the Schema that produced the change.
        """

        return change.get_schema_name()



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""An interface for Transformers that operate on single Items with no metadata needs."""


from abc import abstractmethod

from autotransform.batcher.base import Batch
from autotransform.item.base import Item
from autotransform.transformer.base import Transformer


class SingleTransformer(Transformer[None]):
    """A simple interface for writing a Transformer that operates on an individual Item level."""

    @abstractmethod
    def _transform_item(self, item: Item) -> None:
        """Executes a transformation on a single Item.

        Args:
            item (Item): The Item that is being transformed.
        """

    def transform(self, batch: Batch) -> None:
        """Splits out all Items to be transformed.

        Args:
            batch (Batch): The Batch being transformed.
        """

        for item in batch["items"]:
            self._transform_item(item)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The logging level for an Event. Determines how an event is handled, such
as if it is written to console.
"""


from enum import Enum


class LoggingLevel(int, Enum):
    """The level of an Event, used to determine how it is handled."""

    ERROR = 0
    WARNING = 1
    INFO = 2
    VERBOSE = 3
    DEBUG = 4



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the AggregateCondition."""

from __future__ import annotations

from typing import Any, ClassVar, Dict, List, Type

from autotransform.change.base import Change
from autotransform.step.condition.base import FACTORY as condition_factory
from autotransform.step.condition.base import Condition, ConditionName
from autotransform.util.enums import AggregatorType


class AggregateCondition(Condition):
    """A Condition which aggregates a list of Conditions using the supplied aggregator and
    returns the result of the aggregation.

    Attributes:
        aggregator (AggregatorType): How to aggregate the conditions, using any or all.
        conditions (List[Condition]): The conditions to be aggregated.
        name (ClassVar[ConditionName]): The name of the Component.
    """

    aggregator: AggregatorType
    conditions: List[Condition]

    name: ClassVar[ConditionName] = ConditionName.AGGREGATE

    def check(self, change: Change) -> bool:
        """Checks whether the aggregation of all conditions passes.

        Args:
            change (Change): The Change the Condition is checking.

        Returns:
            bool: Whether the Change passes the Condition.
        """

        if self.aggregator == AggregatorType.ALL:
            return all(condition.check(change) for condition in self.conditions)

        if self.aggregator == AggregatorType.ANY:
            return any(condition.check(change) for condition in self.conditions)

        raise ValueError(f"Unknown aggregator type {self.aggregator}")

    @classmethod
    def from_data(cls: Type[AggregateCondition], data: Dict[str, Any]) -> AggregateCondition:
        """Produces an instance of the component from decoded data.

        Args:
            data (Dict[str, Any]): The JSON decoded data.

        Returns:
            AggregateCondition: An instance of the component.
        """

        aggregator = (
            data["aggregator"]
            if isinstance(data["aggregator"], AggregatorType)
            else AggregatorType(data["aggregator"])
        )
        conditions = [condition_factory.get_instance(condition) for condition in data["conditions"]]
        return cls(aggregator=aggregator, conditions=conditions)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The base class and associated classes for Transformer components."""

from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import Any, ClassVar, Generic, Mapping, Optional, TypeVar

from autotransform.batcher.base import Batch
from autotransform.util.component import ComponentFactory, ComponentImport, NamedComponent

TResult = TypeVar("TResult", bound=Optional[Mapping[str, Any]])


class TransformerName(str, Enum):
    """A simple enum for mapping."""

    JSCODESHIFT = "jscodeshift"
    LIBCST = "libcst"
    OPEN_AI = "open_ai"
    REGEX = "regex"
    SCRIPT = "script"


class Transformer(Generic[TResult], NamedComponent):
    """The base for Transformer components. Transformers are used to execute changes to a codebase.
    A Transformer takes in a Batch and then executes all changes associated with the Batch.

    Attributes:
        name (ClassVar[TransformerName]): The name of the component.
    """

    name: ClassVar[TransformerName]

    @abstractmethod
    def transform(self, batch: Batch) -> TResult:
        """Execute a transformation against the provided Batch. All writing should be done via
        CachedFile's write_content method or FileItem's write_content method to ensure operations
        are easily accessible to testing and file content cache's are kept accurate.

        Args:
            batch (Batch): The Batch that will be transformed.
        """


FACTORY = ComponentFactory(
    {
        TransformerName.JSCODESHIFT: ComponentImport(
            class_name="JSCodeshiftTransformer", module="autotransform.transformer.jscodeshift"
        ),
        TransformerName.LIBCST: ComponentImport(
            class_name="LibCSTTransformer", module="autotransform.transformer.libcst"
        ),
        TransformerName.OPEN_AI: ComponentImport(
            class_name="OpenAITransformer", module="autotransform.transformer.openai"
        ),
        TransformerName.REGEX: ComponentImport(
            class_name="RegexTransformer", module="autotransform.transformer.regex"
        ),
        TransformerName.SCRIPT: ComponentImport(
            class_name="ScriptTransformer", module="autotransform.transformer.script"
        ),
    },
    Transformer,  # type: ignore [type-abstract]
    "transformer.json",
)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""A base for sharding filters used by schedule runs to shard Input items."""

from __future__ import annotations

from abc import abstractmethod

from autotransform.filter.base import Filter
from autotransform.item.base import Item


class ShardFilter(Filter):
    """A base for sharding filters that checks that an Item fits the current valid shard.

    Attributes:
        num_shards (int): The number of shards to split the items across.
        valid_shard (int): The current valid shard to use.
    """

    num_shards: int
    valid_shard: int = -1

    @abstractmethod
    def _shard(self, item: Item) -> int:
        """Produces a shard from an item.

        Args:
            item (Item): The Item to produce a shard number for.

        Returns:
            int: The shard number for the Item.
        """

    def _is_valid(self, item: Item) -> bool:
        """Check whether the item's shard matches the current shard.

        Args:
            item (Item): The Item to check.

        Returns:
            bool: Returns True if the Item's shard matches the current valid shard.
        """

        assert self.valid_shard >= 0, "Shard filter not initialized correctly"
        return self._shard(item) == self.valid_shard



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""A base class for configuration fetching. Defines the API for fetching configuration
so that different components can be used that store configuration in different ways."""

from abc import abstractmethod
from enum import Enum

from autotransform.config.config import Config


class ConfigFetcherName(str, Enum):
    """A simple enum for mapping."""

    DEFAULT = "default"
    ENVIRONMENT = "environment"


class ConfigFetcher:  # pylint: disable=too-few-public-methods
    """An object representing the API needed for config fetching."""

    @abstractmethod
    def get_config(self) -> Config:
        """Fetch the Config.

        Returns:
            Config: The Config for AutoTransform.
        """



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""A script to perform a migration of a Manager JSON file from 1.0.3 to 1.0.5."""

import json
from argparse import ArgumentParser
from typing import Any, Dict, List

from autotransform.config import get_repo_config_relative_path
from autotransform.step.condition.base import ConditionName
from autotransform.util.manager import Manager


def get_arg_parser() -> ArgumentParser:
    """Gets the argument parser for migrating the Manager JSON file.

    Returns:
        ArgumentParser: The arg parser with all args set up.
    """

    parser = ArgumentParser(
        description="Upgrade Manager JSON files for 1.0.0 -> 1.0.1",
        prog="AutoTransform",
    )

    parser.add_argument(
        "--path",
        metavar="path",
        required=False,
        type=str,
        help="A file path to the JSON encoded file, only use if file is in a non-usual place.",
    )

    return parser


def main():
    """Migrate the Manager for 1.0.3 -> 1.0.5."""

    parser = get_arg_parser()
    args = parser.parse_args()
    file_path = args.path
    if file_path is None:
        file_path = f"{get_repo_config_relative_path()}/manager.json"

    with open(file_path, "r", encoding="UTF-8") as manager_file:
        manager_json = manager_file.read()

    manager_data = json.loads(manager_json)
    update_manager_data(manager_data)
    manager = Manager.from_data(manager_data)
    manager.write(file_path)


def update_manager_data(manager_data: Dict[str, Any]) -> None:
    """Updates the Manager data for the new format.

    Args:
        manager_data (Dict[str, Any]): The Manager data to update
    """

    for step in manager_data["steps"]:
        update_step_data(step)


def update_step_data(step_data: Dict[str, Any]) -> None:
    """Updates the Step data to the new format.

    Args:
        step_data (Dict[str, Any]): The Step data to update.
    """

    if step_data["name"] == "conditional":
        update_condition_data(step_data["condition"])


def update_condition_data(condition_data: Dict[str, Any]) -> None:
    """Updates the Condition data to the new format.

    Args:
        condition_data (Dict[str, Any]): The Condition data to update.
    """

    if condition_data["name"] != ConditionName.CHANGE_STATE:
        return

    if isinstance(condition_data["value"], str) and condition_data["value"] in [
        "approved",
        "changes_requested",
    ]:
        condition_data["name"] = ConditionName.REVIEW_STATE
        return

    if isinstance(condition_data["value"], List):
        print("Can not migrate in/not_in comparisons for conditions")


if __name__ == "__main__":
    main()



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the CodeownersFilter."""

from __future__ import annotations

from functools import cached_property
from typing import ClassVar, Optional

from autotransform.filter.base import Filter, FilterName
from autotransform.item.base import Item
from autotransform.item.file import FileItem
from codeowners import CodeOwners


class CodeownersFilter(Filter):
    """A filter which uses Github CODEOWNERS files to separate changes by owner. Titles will
    be of the form 'prefix <owner>'

    Attributes:
        codeowners_location (str): The location of the CODEOWNERS file.
        owner (Optional[str]): The owner to allow files for. If None is provided, checks
            for unowned.
        name (ClassVar[FilterName]): The name of the Component.
    """

    codeowners_location: str
    owner: Optional[str]

    name: ClassVar[FilterName] = FilterName.CODEOWNERS

    @cached_property
    def _owners(self) -> CodeOwners:
        """Gets the parsed CodeOwners as a cached property.

        Returns:
            CodeOwners: The parsed CodeOwners.
        """

        with open(self.codeowners_location, mode="r", encoding="UTF-8") as codeowners_file:
            return CodeOwners(codeowners_file.read())

    @cached_property
    def _formated_owner(self) -> Optional[str]:
        if self.owner is None:
            return None
        return self.owner.removeprefix("@")

    def _is_valid(self, item: Item) -> bool:
        """Checks whether the Item is a file owned by the supplied owner. If no owner
        is supplied, checks for whether the file is unowned.

        Args:
            item (Item): The Item the Filter is checking.

        Returns:
            bool: Whether the Item passes the Filter.
        """

        if not isinstance(item, FileItem):
            return False
        owners = self._owners.of(item.get_path())

        if self.owner is None and not owners:
            return True

        if self.owner is None:
            return False

        for owner in owners:
            if self._formated_owner == owner[1].removeprefix("@"):
                return True

        return False



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""All Actions associated with handling source control options for a Change."""

from typing import ClassVar

from autotransform.change.base import Change
from autotransform.runner.base import Runner
from autotransform.step.action.base import Action, ActionName


class AbandonAction(Action):
    """Abandons an outstanding Change.

    Attributes:
        name (ClassVar[ActionName]): The name of the component.
    """

    name: ClassVar[ActionName] = ActionName.ABANDON

    def run(self, change: Change) -> bool:
        """Abandons a specified Change.

        Args:
            change (Change): The Change to abandon.

        Returns:
            bool: Whether the Change was abandoned successfully.
        """

        return change.abandon()


class MergeAction(Action):
    """Merges an outstanding Change.

    Attributes:
        name (ClassVar[ActionName]): The name of the component.
    """

    name: ClassVar[ActionName] = ActionName.MERGE

    def run(self, change: Change) -> bool:
        """Merges a specified Change.

        Args:
            change (Change): The Change to merge.

        Returns:
            bool: Whether the Change was merged successfully.
        """

        return change.merge()


class NoneAction(Action):
    """Performs no task.

    Attributes:
        name (ClassVar[ActionName]): The name of the component.
    """

    name: ClassVar[ActionName] = ActionName.NONE

    def run(self, change: Change) -> bool:
        """Performs no action.

        Args:
            change (Change): The Change to having no action taken on it

        Returns:
            bool: Always returns True.
        """

        return True


class UpdateAction(Action):
    """Updates an outstanding Change.

    Attributes:
        _runner (ClassVar[Runner]): The runner to use for updating
        name (ClassVar[ActionName]): The name of the component.
    """

    _runner: ClassVar[Runner]

    name: ClassVar[ActionName] = ActionName.UPDATE

    @staticmethod
    def set_runner(runner: Runner) -> None:
        """Sets the runner for the Action to use.

        Args:
            runner (Runner): The Runner to use to update Changes.
        """

        UpdateAction._runner = runner

    def run(self, change: Change) -> bool:
        """Updates a specified Change.

        Args:
            change (Change): The Change to update.

        Returns:
            bool: Whether the Change was updated successfully.
        """

        return change.update(UpdateAction._runner)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the GithubRunner."""

from __future__ import annotations

import json
from typing import Any, ClassVar, Dict, Optional

from autotransform.change.base import Change
from autotransform.event.debug import DebugEvent
from autotransform.event.handler import EventHandler
from autotransform.event.remoterun import RemoteRunEvent
from autotransform.event.update import RemoteUpdateEvent
from autotransform.event.verbose import VerboseEvent
from autotransform.filter.shard import ShardFilter
from autotransform.repo.github import GithubRepo
from autotransform.runner.base import Runner, RunnerName
from autotransform.schema.schema import AutoTransformSchema
from autotransform.util.github import GithubUtils


class GithubRunner(Runner):
    """A Runner component that is used to trigger Github workflows. See
    examples/workflows/autotransform.run.yml.

    Attributes:
        run_workflow (str): The name of the workflow to use for running a Schema.
        update_workflow (str): The name of the workflow to use for updating a Change.
        repo_name (optional, Optional[str]): The name of the Github repo to trigger actions from.
            If not provided, the repo of the Schema will be used. Defaults to None.
        repo_ref (optional, Optional[str]): The ref of the Github repo to trigger actions from.
            If not provided, the base branch of the Schema's repo will be used. Defaults to None.
        target_repo_name (optional, Optional[str]): The name of the Github repo to checkout.
            If not provided, the workflow is expected to define the repo on it's own.
            Defaults to None.
        target_repo_ref (optional, Optional[str]): The ref of the Github repo to checkout.
            If not provided, the workflow is expected to define the repo on it's own.
            Defaults to None.
        name (ClassVar[RunnerName]): The name of the component.
    """

    run_workflow: str
    update_workflow: str

    repo_name: Optional[str] = None
    repo_ref: Optional[str] = None
    target_repo_name: Optional[str] = None
    target_repo_ref: Optional[str] = None

    name: ClassVar[RunnerName] = RunnerName.GITHUB

    def run(self, schema: AutoTransformSchema) -> None:
        """Triggers a full run of a Schema by submitting a workflow run to the
        Github repo in the Schema.

        Args:
            schema (AutoTransformSchema): The schema that will be run.
        """

        event_handler = EventHandler.get()

        # Set up inputs
        workflow_inputs = {"schema": schema.config.schema_name}
        if schema.config.max_submissions:
            workflow_inputs["max_submissions"] = str(schema.config.max_submissions)
        shard_filter = [filt for filt in schema.filters if isinstance(filt, ShardFilter)]
        if shard_filter:
            workflow_inputs["filter"] = json.dumps(shard_filter[0].bundle())

        # Dispatch a Workflow run
        workflow_url = self._create_workflow_dispatch(self.run_workflow, schema, workflow_inputs)
        assert workflow_url is not None, "Failed to dispatch workflow request"

        event_handler.handle(VerboseEvent({"message": "Successfully dispatched workflow run"}))
        if workflow_url == "":
            event_handler.handle(DebugEvent({"message": "No guess for workflow run URL."}))
            return
        event_handler.handle(
            DebugEvent(
                {
                    "message": "Because Github REST API does not provide IDs in response, "
                    + "taking best guess at workflow URL"
                }
            )
        )
        event_handler.handle(
            RemoteRunEvent({"schema_name": schema.config.schema_name, "ref": workflow_url})
        )

    def update(self, change: Change) -> None:
        """Triggers an update of the Change by submitting a workflow run to the
        Github repo in the Schema associated with the change.

        Args:
            change (Change): The Change to update.
        """

        event_handler = EventHandler.get()

        # Dispatch a Workflow run
        workflow_url = self._create_workflow_dispatch(
            self.update_workflow, change.get_schema(), {"change": json.dumps(change.bundle())}
        )
        assert workflow_url is not None, "Failed to dispatch workflow request"

        event_handler.handle(VerboseEvent({"message": "Successfully dispatched workflow run"}))
        if workflow_url == "":
            event_handler.handle(DebugEvent({"message": "No guess for workflow run URL."}))
            return
        event_handler.handle(
            DebugEvent(
                {
                    "message": "Because Github REST API does not provide IDs in response, "
                    + "taking best guess at workflow URL"
                }
            )
        )
        event_handler.handle(RemoteUpdateEvent({"change": change, "ref": workflow_url}))

    def _create_workflow_dispatch(
        self, workflow_name: str, schema: AutoTransformSchema, inputs: Dict[str, Any]
    ) -> str:
        """Creates a workflow dispatch

        Args:
            workflow_name (str): The name of the workflow to dispatch.
            schema (AutoTransformSchema): The Schema that is involved with the dispatch.
            inputs (Dict[str, Any]): The inputs for the dispatch.

        Returns:
            str: The URL for the workflow run.
        """

        if self.repo_name is not None:
            repo_name = self.repo_name
        else:
            repo = schema.repo
            assert isinstance(
                repo, GithubRepo
            ), "GithubRunner can only run using schemas that have Github repos"
            repo_name = repo.full_github_name

        if self.repo_ref is not None:
            repo_ref = self.repo_ref
        else:
            repo = schema.repo
            assert isinstance(
                repo, GithubRepo
            ), "GithubRunner can only run using schemas that have Github repos"
            repo_ref = repo.base_branch

        # Allow controlling the target repo with the Runner
        if self.target_repo_name is not None:
            inputs["target_repo_name"] = self.target_repo_name
        if self.target_repo_ref is not None:
            inputs["target_repo_ref"] = self.target_repo_ref

        # Dispatch a Workflow run
        workflow_url = GithubUtils.get(repo_name).create_workflow_dispatch(
            workflow_name,
            repo_ref,
            inputs,
        )
        assert workflow_url is not None, "Failed to dispatch workflow request"
        return workflow_url



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the CreatedAgoCondition."""

from __future__ import annotations

import time
from typing import ClassVar, List

from autotransform.change.base import Change
from autotransform.step.condition.base import ConditionName, SortableComparisonCondition
from autotransform.step.condition.comparison import ComparisonType


class CreatedAgoCondition(SortableComparisonCondition[int]):
    """A condition which checks how long ago a Change was created against the supplied time, all
    in seconds, using the supplied comparison.

    Attributes:
        comparison (ComparisonType): The type of comparison to perform.
        value (int | List[int]): The number of seconds to compare against.
        name (ClassVar[ConditionName]): The name of the Component.
    """

    comparison: ComparisonType
    value: int | List[int]

    name: ClassVar[ConditionName] = ConditionName.CREATED_AGO

    def get_val_from_change(self, change: Change) -> int:
        """Gets how long ago the Change was created.

        Args:
            change (Change): The Change the Condition is checking.

        Returns:
            int: How long ago the Change was created.
        """

        return int(time.time() - change.get_created_timestamp())



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Utility methods for getting user input for AutoTransform."""

from getpass import getpass
from typing import List, Optional, Tuple, TypeVar

from colorama import Fore

T = TypeVar("T")

ERROR_COLOR = Fore.RED
INFO_COLOR = Fore.YELLOW
INPUT_COLOR = Fore.GREEN
RESET_COLOR = Fore.RESET


def info(text: str) -> None:
    """Prints a string of text as info to the console.

    Args:
        text (str): The text to print.
    """

    print(f"{INFO_COLOR}{text}{RESET_COLOR}")


def error(text: str) -> None:
    """Prints a string of text as error to the console.

    Args:
        text (str): The text to print.
    """

    print(f"{ERROR_COLOR}{text}{RESET_COLOR}")


def get_str(prompt: str, secret: bool = False) -> str:
    """Prompts the user to input a value.

    Args:
        prompt (str): The prompt to give to the user
        secret (bool, optional): Whether to use getpass for input. Defaults to False.

    Returns:
        str: The input value.
    """

    if secret:
        return getpass(f"{INPUT_COLOR}{prompt}{RESET_COLOR}")
    return input(f"{INPUT_COLOR}{prompt}{RESET_COLOR}")


def input_string(
    prompt: str,
    name: str,
    previous: Optional[str] = None,
    default: Optional[str] = None,
    secret: bool = False,
) -> str:
    """Prompts the user to input a value, or potentially use a previously input value/default value.

    Args:
        prompt (str): The prompt to give to the user.
        name (str): The name of the value being prompted for.
        previous (Optional[str], optional): The previously input value. Defaults to None.
        default (Optional[str], optional): The default value. Defaults to None.
        secret (bool, optional): Whether to use getpass for inputs. Defaults to False.

    Returns:
        str: The value entered by the user.
    """

    # Check if the previous value should be used
    # Ignore previous value if it is the same as default
    if (
        previous is not None
        and previous != default
        and choose_yes_or_no(f"Use previous {name} ({previous})?")
    ):
        return previous

    # Check if the user wants to use the default value
    if default is not None and choose_yes_or_no(f"Use default {name}?"):
        return default

    return get_str(f"{prompt} ", secret)


# pylint: disable=too-many-branches
def input_ints(
    prompt: str,
    min_val: Optional[int] = None,
    max_val: Optional[int] = None,
    min_choices: int = 1,
    max_choices: int = 1,
) -> List[int]:
    """Gets integers input from the user within the specified range.

    Args:
        prompt (str): The prompt to give the user.
        min_val (Optional[int], optional): The minimum acceptable value. Defaults to None.
        max_val (Optional[int], optional): The maximum acceptable value. Defaults to None.
        min_choices(int, optional): The minimum number of selections to be made. Defaults to 1.
        max_choices(int, optional): The maximum number of selections to be made. Defaults to 1.

    Returns:
        List[int]: The integers specified by the user.
    """

    assert min_choices >= 0
    assert max_choices >= min_choices
    if min_val is not None and max_val is not None:
        assert min_val <= max_val, "The minimum valid value must be less than the maximum"
        assert max_val - min_val + 1 >= max_choices
        if max_val - min_val + 1 == min_choices:
            return list(range(min_val, max_val + 1))
        range_str = f"({min_val}-{max_val})"
    elif min_val is not None:
        range_str = f"(>={min_val})"
    elif max_val is not None:
        range_str = f"(<={max_val})"
    else:
        range_str = ""

    if max_choices > 1 and min_choices == 0:
        choice_str = "(separate selections with commas, blank to choose none)"
    elif max_choices > 1:
        choice_str = "(separate selections with commas)"
    elif min_choices == 0:
        choice_str = "(blank to choose none)"
    else:
        choice_str = ""

    while True:
        str_val = get_str(f"{prompt}{range_str}{choice_str}: ")
        if max_choices > 1:
            vals = [val.strip() for val in str_val.split(",")]
        else:
            vals = [str_val.strip()]
        int_vals = []
        if min_choices == 0:
            vals = [val for val in vals if val != ""]
        invalid_value = False
        for val in vals:
            if val.startswith("-"):
                is_negative = True
                val = val[1:]
            else:
                is_negative = False
            if not val.isdigit():
                error("Only a decimal number may be entered")
                invalid_value = True
                break
            int_val = -int(val) if is_negative else int(val)
            if min_val is not None and int_val < min_val:
                error(f"{int_val} is too low, must be at least {min_val}")
                invalid_value = True
                break
            if max_val is not None and int_val > max_val:
                error(f"{int_val} is too high, must be less than {max_val}")
                invalid_value = True
                break
            int_vals.append(int_val)
        if invalid_value:
            continue
        if len(vals) > max_choices:
            error(f"Too many selections, only {max_choices} allowed: {vals}")
            continue
        if len(vals) < min_choices:
            error(f"Too few selections, at least {min_choices} required: {vals}")
            continue
        return int_vals


def input_int(
    prompt: str,
    min_val: Optional[int] = None,
    max_val: Optional[int] = None,
) -> int:
    """Gets an integer input from the user within the specified range.

    Args:
        prompt (str): The prompt to give the user.
        min_val (Optional[int], optional): The minimum acceptable value. Defaults to None.
        max_val (Optional[int], optional): The maximum acceptable value. Defaults to None.

    Returns:
        int: The integer specified by the user.
    """

    return input_ints(prompt, min_val=min_val, max_val=max_val)[0]


def choose_option(prompt: str, options: List[Tuple[T, List[str]]]) -> T:
    """Prompts the user to choose one of a set of options using a string. User input is
    converted in to lower case.

    Args:
        prompt (str): The prompt to give the user.
        options (List[Tuple[T, List[str]]]): The potential options to choose. The first value
            is the option, the second value is a list of potential aliases for the option.

    Returns:
        str: The selected option
    """

    assert len(options) > 0, "Choosing from an empty list of options is not possible."
    option_list = [option[1][0] for option in options]
    while True:
        choice = get_str(f"{prompt}({'/'.join(option_list)}) ").lower()
        for option in options:
            if choice in option[1]:
                return option[0]

        error(f"Invalid choice, choose one of: {', '.join(option_list)}")


def choose_yes_or_no(prompt: str) -> bool:
    """Gives the user a yes or no prompt.

    Args:
        prompt (str): The prompt to give to the user.

    Returns:
        bool: If the user chose yes.
    """

    return choose_option(prompt, [(True, ["yes", "y"]), (False, ["no", "n"])])


def choose_options_from_list(
    prompt: str, options: List[Tuple[T, str]], min_choices: int = 1, max_choices: int = 1
) -> List[T]:
    """Prompts the user to choose one of a set of options using a list
    and having the user choose a number.

    Args:
        prompt (str): The prompt for the user.
        options (List[Tuple[T, str]]): The options to choose from, where the first value
            in a Tuple is the option and the second value is the prompt for the list.
        min_choices(int, optional): The minimum number of selections to be made. Defaults to 1.
        max_choices(int, optional): The maximum number of selections to be made. Defaults to 1.

    Returns:
        List[T]: The chosen options.
    """

    assert len(options) > 0, "Choosing from an empty list of options is not possible."
    for idx, option in enumerate(options):
        print(f"{INPUT_COLOR}{idx + 1}) {option[1]}{RESET_COLOR}")
    choices = input_ints(
        prompt,
        min_val=1,
        max_val=len(options),
        min_choices=min_choices,
        max_choices=max_choices,
    )
    return [options[choice - 1][0] for choice in choices]



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation of Request based conditions."""

import os
from functools import cached_property
from typing import Any, ClassVar, List, Mapping, Optional, TypeVar

from pydantic import Field

from autotransform.change.base import Change
from autotransform.step.condition.base import ComparisonCondition, ConditionName
from autotransform.step.condition.comparison import ComparisonType
from autotransform.util.request import RequestHandler

T = TypeVar("T")


class RequestStrCondition(ComparisonCondition[str]):
    """Performs a URL request to get a check a condition on a Change. Used for interacting with
    REST APIs. Handles the basics of the condition.

    Attributes:
        comparison (ComparisonType): The type of comparison to perform.
        url(str): The URL to send a request to.
        value (str | List[str]): The value to compare against.
        data(optional, Mapping[str, Any]): Data to include in the request. Defaults to {}.
        headers(optional, Mapping[str, Any]): Headers to include in the request. Defaults to {}.
        log_response(optional, bool): Indicates whether to log the response using DebugEvent.
            Defaults to False.
        params(optional, Mapping[str, Any]): Params to include in the request. Defaults to {}.
        post(optional, bool): Whether to send the request as a POST. Defaults to True.
        response_field(optional, Optional[str]): The field containing the desired response value
            when using the response as JSON. If multiple levels are needed, they should be separated
            by //. 'foo//bar' as an example would get the field at response['foo']['bar']. If no
            response_field is provided the whole response as text is used. Defaults to None.
        name (ClassVar[ConditionName]): The name of the Component.
    """

    comparison: ComparisonType
    url: str
    value: str | List[str]

    data: Mapping[str, Any] = Field(default_factory=dict)
    headers: Mapping[str, Any] = Field(default_factory=dict)
    log_response: bool = False
    params: Mapping[str, Any] = Field(default_factory=dict)
    post: bool = True
    response_field: Optional[str] = None

    name: ClassVar[ConditionName] = ConditionName.REQUEST_STR

    @cached_property
    def _handler(self) -> RequestHandler:
        """Gets the handler for requests.

        Returns:
            RequestHandler: The handler for requests.
        """

        return RequestHandler(
            url=self.url,
            data=self.data,
            headers=self.headers,
            params=self.params,
            log_response=self.log_response,
            post=self.post,
            constant_replacers={"env": lambda var: str(os.getenv(var) or "")},
        )

    def get_val_from_change(self, change: Change) -> str:
        """Gets the existing value to compare against from the Change.

        Args:
            change (Change): The Change the Condition is checking.

        Returns:
            str: The value from the Change to compare against.
        """

        response = self._handler.get_response(
            ({"change": lambda name: str(getattr(change, name) or "")})
        )

        if not self.response_field:
            return response.text

        result = response.json()
        for field_name in self.response_field.split("//"):
            result = result[field_name]

        return str(result)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the FileItem."""

from typing import ClassVar

from autotransform.item.base import Item, ItemName
from autotransform.util.cachedfile import CachedFile


class FileItem(Item):
    """An item representing a file that can be used by AutoTransform. Provides some utility methods
    for interacting with the file.

    Attributes:
        name (ClassVar[ItemName]): The name of the component.
    """

    name: ClassVar[ItemName] = ItemName.FILE

    def get_path(self) -> str:
        """Gets the path of the file.

        Returns:
            str: The file's path.
        """

        return self.key

    def get_content(self) -> str:
        """Gets the content of the file.

        Returns:
            str: The file's content.
        """

        return CachedFile(self.get_path()).get_content()

    def write_content(self, content: str) -> None:
        """Writes new content to the file.

        Args:
            content (str): The new content for the File.
        """

        CachedFile(self.get_path()).write_content(content)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the AggregateFilter."""

from __future__ import annotations

from typing import Any, ClassVar, Dict, Sequence, Type

from autotransform.filter.base import FACTORY as filter_factory
from autotransform.filter.base import Filter, FilterName
from autotransform.item.base import Item
from autotransform.step.condition.aggregate import AggregatorType


class AggregateFilter(Filter):
    """A Filter which aggregates a list of Filters using the supplied aggregator and
    returns the result of the aggregation.

    Attributes:
        aggregator (AggregatorType): How to aggregate the filters, using any or all.
        filters (Sequence[Filter]): The filters to be aggregated.
        name (ClassVar[FilterName]): The name of the Component.
    """

    aggregator: AggregatorType
    filters: Sequence[Filter]

    name: ClassVar[FilterName] = FilterName.AGGREGATE

    def _is_valid(self, item: Item) -> bool:
        """Checks whether the aggregation of all filters passes.

        Args:
            item (Item): The Item the Filter is checking.

        Returns:
            bool: Whether the Item passes the Filter.
        """

        if self.aggregator == AggregatorType.ALL:
            return all(filter.is_valid(item) for filter in self.filters)

        if self.aggregator == AggregatorType.ANY:
            return any(filter.is_valid(item) for filter in self.filters)

        raise ValueError(f"Unknown aggregator type {self.aggregator}")

    @classmethod
    def from_data(cls: Type[AggregateFilter], data: Dict[str, Any]) -> AggregateFilter:
        """Produces an instance of the component from decoded data.

        Args:
            data (Dict[str, Any]): The JSON decoded data.

        Returns:
            AggregateFilter: An instance of the component.
        """

        aggregator = (
            data["aggregator"]
            if isinstance(data["aggregator"], AggregatorType)
            else AggregatorType(data["aggregator"])
        )
        filters = [filter_factory.get_instance(filter) for filter in data["filters"]]
        return cls(aggregator=aggregator, filters=filters)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the ScriptValidator."""

from __future__ import annotations

from typing import Any, ClassVar, List, Mapping, Optional, Sequence

import autotransform.schema
from autotransform.batcher.base import Batch
from autotransform.event.handler import EventHandler
from autotransform.event.verbose import VerboseEvent
from autotransform.item.base import Item
from autotransform.item.file import FileItem
from autotransform.util.functions import run_cmd_on_items
from autotransform.validator.base import (
    ValidationResult,
    ValidationResultLevel,
    Validator,
    ValidatorName,
)


class ScriptValidator(Validator):
    """Runs a script with the supplied arguments to perform validation. If the per_item flag is
    set to True, the script will be invoked on each Item. If run_on_changes is set to True, the
    script will replace the Batch Items with FileItems for each changed file. The failure_level
    indicates the result level if the script returns a non-zero exit code. Sentinel values can
    be used in args to provide custom arguments for a run.
    The available sentinel values for args are:
        <<KEY>>: A json encoded list of the Items for a Batch. If the per_item flag is set
            this will simply be the key of an Item.
        <<EXTRA_DATA>>: A JSON encoded mapping from Item key to that Item's extra_data. If the
            per_item flag is set, this will simply be a JSON encoding of the Item's extra_data.
            If extra_data is not present for an item, it is treated as an empty Dict.
        <<METADATA>>: A JSON encoded version of the Batch's metadata.
    _FILE can be appended to any of these (i.e. <<KEY_FILE>>) and the arg will instead be replaced
     with a path to a file containing the value.

    Attributes:
        args (List[str]): The arguments to supply to the script.
        script (str): The script to run.
        failure_level (ValidationResultLevel, optional): The result level to use if validation
            fails. Defaults to ValidationResultLevel.ERROR.
        per_item (bool, optional): Whether to run the script on each item. Defaults to False.
        run_on_changes (bool, optional): Whether to replace the Items in the batch with
            FileItems for the changed files. Defaults to False.
        name (ClassVar[ValidatorName]): The name of the Component.
    """

    args: List[str]
    script: str
    failure_level: ValidationResultLevel = ValidationResultLevel.ERROR
    per_item: bool = False
    run_on_changes: bool = False

    name: ClassVar[ValidatorName] = ValidatorName.SCRIPT

    def check(self, batch: Batch, _transform_data: Optional[Mapping[str, Any]]) -> ValidationResult:
        """Runs the script validation against the Batch, either on each item individually or
        on the entire Batch, based on the per_item flag. If the script returns a non-zero exit
        code, the failure_level will be in the result.

        Args:
            batch (Batch): The transformed Batch to validate.
            _transform_data (Optional[Mapping[str, Any]]): Data from the transformation. Unused.

        Returns:
            ValidationResult: The result of the validation check indicating the severity of any
                validation failures as well as an associated message
        """
        if not self.per_item:
            return self._check_batch(batch)
        if self.run_on_changes:
            current_schema = autotransform.schema.current
            assert current_schema is not None
            repo = current_schema.repo
            assert repo is not None
            items: Sequence[Item] = [FileItem(key=file) for file in repo.get_changed_files(batch)]
        else:
            items = batch["items"]

        for item in items:
            result = self._check_single(item, batch.get("metadata", None))
            if result.level != ValidationResultLevel.NONE:
                return result
        return ValidationResult(level=ValidationResultLevel.NONE, validator=self)

    def _check_single(
        self, item: Item, batch_metadata: Optional[Mapping[str, Any]]
    ) -> ValidationResult:
        """Executes a simple script to validate a single Item.

        Args:
            item (Item): The Item that will be validated.
            batch_metadata (Optional[Mapping[str, Any]]): The metadata of the Batch containing the
                Item.
        """

        event_handler = EventHandler.get()

        # Get Command
        cmd = [self.script]
        cmd.extend(self.args)

        proc = run_cmd_on_items(cmd, [item], batch_metadata or {})

        # Handle output
        level = self.failure_level if proc.returncode != 0 else ValidationResultLevel.NONE
        if proc.stdout.strip() != "":
            event_handler.handle(VerboseEvent({"message": f"STDOUT:\n{proc.stdout.strip()}"}))
        else:
            event_handler.handle(VerboseEvent({"message": "No STDOUT"}))
        if proc.stderr.strip() != "":
            event_handler.handle(VerboseEvent({"message": f"STDERR:\n{proc.stderr.strip()}"}))
        else:
            event_handler.handle(VerboseEvent({"message": "No STDERR"}))
        return ValidationResult(
            level=level,
            message=f"[{cmd}]\nSTDOUT:\n{proc.stdout.strip()}\nSTDERR:\n{proc.stderr.strip()}",
            validator=self,
        )

    def _check_batch(self, batch: Batch) -> ValidationResult:
        """Executes a simple script to validate the given Batch.

        Args:
            batch (Batch): The batch that will be validated.
        """

        event_handler = EventHandler.get()

        cmd = [self.script]

        # Get items
        if self.run_on_changes:
            current_schema = autotransform.schema.current
            assert current_schema is not None
            repo = current_schema.repo
            assert repo is not None
            items: Sequence[Item] = [FileItem(key=file) for file in repo.get_changed_files(batch)]
        else:
            items = batch["items"]

        # Get Command
        cmd = [self.script]
        cmd.extend(self.args)

        proc = run_cmd_on_items(cmd, items, batch.get("metadata", {}))

        # Handle output
        level = self.failure_level if proc.returncode != 0 else ValidationResultLevel.NONE
        if proc.stdout.strip() != "":
            event_handler.handle(VerboseEvent({"message": f"STDOUT:\n{proc.stdout.strip()}"}))
        else:
            event_handler.handle(VerboseEvent({"message": "No STDOUT"}))
        if proc.stderr.strip() != "":
            event_handler.handle(VerboseEvent({"message": f"STDERR:\n{proc.stderr.strip()}"}))
        else:
            event_handler.handle(VerboseEvent({"message": "No STDERR"}))
        return ValidationResult(
            level=level,
            message=f"[{cmd}]\nSTDOUT:\n{proc.stdout.strip()}\nSTDERR:\n{proc.stderr.strip()}",
            validator=self,
        )



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The base class and associated classes for SchemaBuilders."""

import json
from abc import abstractmethod
from enum import Enum
from typing import List, Optional

from autotransform.batcher.base import Batcher
from autotransform.batcher.single import SingleBatcher
from autotransform.command.base import Command
from autotransform.filter.base import Filter
from autotransform.input.base import Input
from autotransform.repo.base import Repo
from autotransform.schema.config import SchemaConfig
from autotransform.schema.schema import AutoTransformSchema
from autotransform.transformer.base import Transformer
from autotransform.util.component import ComponentFactory, NamedComponent
from autotransform.validator.base import Validator


class SchemaBuilderName(str, Enum):
    """A simple enum for mapping."""


class SchemaBuilder(NamedComponent):
    """The base for SchemaBuilders. SchemaBuilders are used for programatic Schema generation.
    This can be used in conjunction with inputs or configuration to customize Schemas run
    through automation. Can also be used to generate JSON Schemas that can be utilized.
    """

    @abstractmethod
    def get_input(self) -> Input:
        """Get the Input for the Schema.

        Returns:
            Input: The Input that will be used in the built Schema.
        """

    def get_filters(self) -> List[Filter]:
        """Get the Filters for the Schema.

        Returns:
            List[Filter]: The Filters that will be used in the built Schema.
        """

        return []

    def get_batcher(self) -> Batcher:
        """Get the Batcher for the Schema.

        Returns:
            Batcher: The Batcher that will be used in the built Schema.
        """

        return SingleBatcher(title="")

    @abstractmethod
    def get_transformer(self) -> Transformer:
        """Get the Transformer for the Schema.

        Returns:
            Transformer: The Transformer that will be used in the built Schema.
        """

    def get_validators(self) -> List[Validator]:
        """Get the Validators for the Schema.

        Returns:
            List[Validator]: The Validators that will be used in the built Schema.
        """

        return []

    def get_commands(self) -> List[Command]:
        """Get the Commands for the Schema.

        Returns:
            List[Command]: The Commands that will be used in the built Schema.
        """

        return []

    def get_repo(self) -> Optional[Repo]:
        """Get the Repo for the Schema.

        Returns:
            Repo: The Repo that will be used in the built Schema.
        """

        return None

    def get_config(self) -> SchemaConfig:
        """Get the SchemaConfig for the Schema.

        Returns:
            SchemaConfig: The SchemaConfig that will be used in the built Schema.
        """

        return SchemaConfig(schema_name=type(self).__name__)

    def build(self) -> AutoTransformSchema:
        """Builds a Schema based on the state of the SchemaBuilder.

        Returns:
            AutoTransformSchema: The Schema produced by this SchemaBuilder.
        """

        return AutoTransformSchema(
            input=self.get_input(),
            batcher=self.get_batcher(),
            transformer=self.get_transformer(),
            filters=self.get_filters(),
            validators=self.get_validators(),
            commands=self.get_commands(),
            repo=self.get_repo(),
            config=self.get_config(),
        )

    def dump_to_file(self, path: str) -> None:
        """Dumps the Schema this SchemaBuilder would produce to the file located at the provided
        path.

        Args:
            path (str): The path of the file to dump the JSON encoded Schema to.
        """

        # pylint: disable=unspecified-encoding

        with open(path, "w") as file:
            file.write(json.dumps(self.build().bundle(), indent=4))


FACTORY = ComponentFactory(
    {},
    SchemaBuilder,  # type: ignore [type-abstract]
    "schema_builder.json",
)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the GithubRepo."""

from __future__ import annotations

import json
from copy import deepcopy
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Mapping, Optional, Sequence

import autotransform.schema
from autotransform.batcher.base import Batch
from autotransform.change.base import Change
from autotransform.change.github import GithubChange
from autotransform.event.handler import EventHandler
from autotransform.event.verbose import VerboseEvent
from autotransform.repo.base import RepoName
from autotransform.repo.git import GitRepo
from autotransform.util.github import GithubUtils
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from pydantic import Field

if TYPE_CHECKING:
    from autotransform.schema.schema import AutoTransformSchema


# pylint: disable=too-few-public-methods
class GithubRepoBatchMetadata(BaseModel):
    """The metadata used by GithubRepo.

    Attributes:
        body (str): The body of the Pull Request to make.
        labels(optional, List[str]): The labels to add to the Pull Request. Defaults to [].
        reviewers(optional, List[str]): The reviewers to request for the Pull Request.
            Defaults to [].
        team_reviewers(optional, List[str]): The team reviewers to request for the Pull Request.
            Defaults to [].
    """

    body: str
    labels: List[str] = Field(default_factory=list)
    reviewers: List[str] = Field(default_factory=list)
    team_reviewers: List[str] = Field(default_factory=list)


class GithubRepo(GitRepo):
    """A Repo that provides support for submitting changes as a pull request against
    a Github repo.

    Attributes:
        full_github_name (str): The fully qualified name of the Github Repo.
        commit_repo (optional, Optional[str]): The repo from which to generate a Pull Request. Used
            if creating a Pull Request from a fork. Defaults to None.
        hide_automation_info (bool, optional): Whether to hide information on how automation was
            done from the pull request body. Defaults to False.
        hide_autotransform_docs (bool, optional): Whether to hide links to AutoTransform docs
            from the pull request body. Defaults to False.
        labels (List[str], optional): The labels to add to pull requests. Defaults to [].
        reviewers (List[str], optional): The reviewers to request for pull requests. Defaults to [].
        team_reviewers (List[str], optional): The team reviewers to request for pull requests.
            Defaults to [].
        name (ClassVar[RepoName]): The name of the component.
    """

    full_github_name: str

    commit_repo: Optional[str] = None
    hide_automation_info: bool = False
    hide_autotransform_docs: bool = False
    labels: List[str] = Field(default_factory=list)
    reviewers: List[str] = Field(default_factory=list)
    team_reviewers: List[str] = Field(default_factory=list)

    name: ClassVar[RepoName] = RepoName.GITHUB

    def has_outstanding_change(self, batch: Batch) -> bool:
        """Checks the state of the repo to see whether an outstanding Change
        for the Batch exists.

        Args:
            batch (Batch): The Batch to check for.

        Returns:
            bool: Whether an outstanding Change exists.
        """

        refs = self._local_repo.git.ls_remote(
            "--heads",
            self._local_repo.remote().name,
            GitRepo.get_branch_name(batch["title"]),
        )
        return refs != ""

    def submit(
        self,
        batch: Batch,
        _transform_data: Optional[Mapping[str, Any]],
        change: Optional[Change] = None,
    ) -> None:
        """Performs the normal submit for a git repo then submits a pull request
        against the provided Github repo.

        Args:
            batch (Batch): The Batch for which the changes were made.
            _transform_data (Optional[Mapping[str, Any]]): Data from the transformation. Unused.
            change (Optional[Change]): An associated change which should be updated.
                Defaults to None.
        """

        title = GitRepo.get_commit_message(batch["title"])
        batch_metadata = GithubRepoBatchMetadata.parse_obj(batch.get("metadata", {}))

        self.commit(batch["title"], change is not None)

        commit_branch = GitRepo.get_branch_name(batch["title"])
        remote = self._local_repo.remote()
        if change is not None:
            self._local_repo.git.push(remote.name, "-u", "-f", commit_branch)
            return

        self._local_repo.git.push(remote.name, "-u", commit_branch)

        if self.hide_automation_info:
            automation_info = ""
        else:
            automation_info = "\n\n" + self.get_automation_info(autotransform.schema.current, batch)

        if self.commit_repo is not None:
            head = f"{self.commit_repo}:"
        else:
            head = ""

        pull_request = GithubUtils.get(self.full_github_name).create_pull_request(
            title,
            f"{str(batch_metadata.body)}{automation_info}",
            self._base_branch.name,
            f"{head}{commit_branch}",
        )

        EventHandler.get().handle(
            VerboseEvent({"message": f"Pull request created: {pull_request.number}"})
        )

        # Add labels
        labels = deepcopy(batch_metadata.labels)
        labels.extend(self.labels)
        if labels:
            pull_request.add_labels(labels)

        # Request reviewers
        reviewers = deepcopy(batch_metadata.reviewers)
        reviewers.extend(self.reviewers)
        team_reviewers = deepcopy(batch_metadata.team_reviewers)
        team_reviewers.extend(self.team_reviewers)
        if reviewers or team_reviewers:
            pull_request.add_reviewers(reviewers, team_reviewers)

    def get_automation_info(self, schema: Optional[AutoTransformSchema], batch: Batch) -> str:
        """Gets information on automating with AutoTransform.

        Args:
            schema (Optional[AutoTransformSchema]): The Schema making the Change.
            batch (Batch): The Batch the Change is being made for.

        Returns:
            str: The text for automating.
        """

        # Create body content with information on replicating the change
        automation_info_lines = ["ADDED AUTOMATICALLY BY AUTOTRANSFORM"]
        if not self.hide_autotransform_docs:
            automation_info_lines.append(
                "Learn more about AutoTransform [here](https://autotransform.readthedocs.io)"
            )
        automation_info_lines.append("Schema and batch information for the change below")

        files = {}
        # Add schema JSON
        if schema is not None:
            files["schema"] = {"content": json.dumps(schema.bundle(), indent=4)}

        # Add batch JSON
        chunk_size = 5000
        items = batch["items"]
        item_chunks = [items[i : i + chunk_size] for i in range(0, len(items), chunk_size)]
        if not item_chunks:
            item_chunks = [[]]
        encodable_batch: Dict[str, Any] = {
            "title": batch["title"],
            "items": [item.bundle() for item in item_chunks[0]],
        }
        if "metadata" in batch:
            encodable_batch["metadata"] = batch["metadata"]
        files["batch"] = {"content": json.dumps(encodable_batch, indent=4)}
        gists = [
            GithubUtils.get(self.full_github_name).create_gist(
                files, description="Automation info for AutoTransform", public=True
            )
        ]
        for i in range(1, len(item_chunks)):
            item_file = {
                "items": {
                    "content": json.dumps([item.bundle() for item in item_chunks[i]], indent=4)
                }
            }
            gists.append(
                GithubUtils.get(self.full_github_name).create_gist(
                    item_file, description="Automation info for AutoTransform", public=True
                )
            )
        automation_info_lines.append(
            f"<<<Automation Info Gist: {'/'.join([gist.gist_id for gist in gists])}>>>"
        )

        return "\n".join(automation_info_lines)

    def get_outstanding_changes(self) -> Sequence[GithubChange]:
        """Gets all outstanding pull requests for the Repo.

        Returns:
            Sequence[GithubChange]: The outstanding Changes against the Repo.
        """

        pulls = GithubUtils.get(self.full_github_name).get_open_pull_requests(self.base_branch)
        authenticated_user_id = GithubUtils.get(self.full_github_name).get_user_id()
        return [
            GithubChange(
                full_github_name=self.full_github_name,
                pull_number=pull.number,
            )
            for pull in pulls
            if pull.owner_id == authenticated_user_id
        ]



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for Conditions based on reviewers."""

from __future__ import annotations

from typing import ClassVar, List, Optional

from autotransform.change.base import Change
from autotransform.step.condition.base import ConditionName, ListComparisonCondition
from autotransform.step.condition.comparison import ComparisonType


class ReviewersCondition(ListComparisonCondition[str]):
    """A condition which checks the reviewers for a Change.

    Attributes:
        comparison (ComparisonType): The type of comparison to perform.
        value (optional, Optional[str]): The reviewer to check for. Defaults to None.
        name (ClassVar[ConditionName]): The name of the Component.
    """

    comparison: ComparisonType
    value: Optional[str] = None

    name: ClassVar[ConditionName] = ConditionName.REVIEWERS

    def get_val_from_change(self, change: Change) -> List[str]:
        """Gets the reviewers from the Change.

        Args:
            change (Change): The Change the Condition is checking.

        Returns:
            List[str]: The reviewers of the Change.
        """

        return change.get_reviewers()


class TeamReviewersCondition(ListComparisonCondition[str]):
    """A condition which checks the team reviewers for a Change.

    Attributes:
        comparison (ComparisonType): The type of comparison to perform.
        value (optional, Optional[str]): The team reviewer to check for. Defaults to None.
        name (ClassVar[ConditionName]): The name of the Component.
    """

    comparison: ComparisonType
    value: Optional[str] = None

    name: ClassVar[ConditionName] = ConditionName.TEAM_REVIEWERS

    def get_val_from_change(self, change: Change) -> List[str]:
        """Gets the team reviewers from the Change.

        Args:
            change (Change): The Change the Condition is checking.

        Returns:
            List[str]: The team reviewers of the Change.
        """

        return change.get_team_reviewers()



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the ExtraDataBatcher."""

from __future__ import annotations

from copy import deepcopy
from typing import Any, ClassVar, Dict, List, Sequence

from pydantic import Field

from autotransform.batcher.base import Batch, Batcher, BatcherName
from autotransform.item.base import Item


class ExtraDataBatcher(Batcher):
    """A Batcher which uses the extra data on Items to create batches.

    Attributes:
        group_by (str): The key of the extra_data on items to group Items by for Batches.
        metadata_keys (optional, List[str]): A list of keys from Items to combine in to the
            metadata of a batch. Defaults to [].

        name (ClassVar[BatcherName]): The name of the Component.
    """

    group_by: str
    metadata_keys: List[str] = Field(default_factory=list)

    name: ClassVar[BatcherName] = BatcherName.EXTRA_DATA

    def batch(self, items: Sequence[Item]) -> List[Batch]:
        """Take filtered Items and group them by an extra_data value.

        Args:
            items (Sequence[Item]): The filtered Items to separate.

        Returns:
            List[Batch]: A list of Batches grouped by the extra_data of the Items.
        """

        groups: Dict[str, List[Item]] = {}
        for item in items:
            extra_data = item.extra_data or {}
            group_by_val = extra_data[self.group_by]
            assert isinstance(group_by_val, str), "Group by values must be strings"
            if group_by_val in groups:
                groups[group_by_val].append(item)
            else:
                groups[group_by_val] = [item]

        batches: List[Batch] = []
        for group_title, group_items in groups.items():
            batch: Batch = {"items": group_items, "title": group_title}
            if self.metadata_keys:
                metadata: Dict[str, List[Any]] = {}
                for key in self.metadata_keys:
                    metadata[key] = []
                for item in group_items:
                    extra_data = item.extra_data or {}
                    for key in self.metadata_keys:
                        val = extra_data.get(key)
                        if isinstance(val, list):
                            metadata[key].extend(deepcopy(val))
                        elif val is not None:
                            metadata[key].append(deepcopy(val))
                for key in self.metadata_keys:
                    metadata[key] = list(set(metadata[key]))
                batch["metadata"] = metadata
            batches.append(batch)
        return batches



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Utilities for handling components, such as importing custom components."""

from __future__ import annotations

import importlib
import json
from abc import ABC
from enum import Enum
from functools import cached_property
from typing import Any, ClassVar, Dict, Generic, List, Optional, Tuple, Type, TypeVar

from autotransform.config import get_config
from autotransform.event.debug import DebugEvent
from autotransform.event.handler import EventHandler
from autotransform.event.warning import WarningEvent
from autotransform.util.console import (
    choose_options_from_list,
    choose_yes_or_no,
    error,
    get_str,
    info,
)
from pydantic import BaseModel  # pylint: disable=no-name-in-module

TComponent = TypeVar("TComponent", bound=BaseModel)

UNSET_VALUE = object()


class ComponentModel(BaseModel):
    """A base class for AutoTransform components that need to handle bundling/unbundling
    for JSON."""

    def bundle(self) -> Dict[str, Any]:
        """Generates a JSON encodable bundle.
        If a component is not JSON encodable this method should be overridden to provide
        an encodable version.

        Returns:
            Dict[str, Any]: The encodable bundle.
        """

        bundle = dict(self._iter(to_dict=False, exclude_defaults=True, exclude_unset=True))
        for key, value in bundle.items():
            if isinstance(value, ComponentModel):
                bundle[key] = value.bundle()
            elif isinstance(value, List) and all(
                isinstance(item, ComponentModel) for item in value
            ):
                bundle[key] = [item.bundle() for item in value]
            elif isinstance(value, Dict) and all(
                isinstance(item, ComponentModel) for item in value.values()
            ):
                bundle[key] = {item_key: item.bundle() for item_key, item in value.items()}

        return bundle

    @classmethod
    def from_data(cls: Type[TComponent], data: Dict[str, Any]) -> TComponent:
        """Produces an instance of the component from decoded data. Override if
        the component had to be modified to encode.

        Args:
            data (Mapping[str, Any]): The JSON decoded data.

        Returns:
            TComponent: An instance of the component.
        """

        return cls.parse_obj(data)

    def __repr__(self) -> str:
        redacted_fields = ["github_token", "jenkins_token"]
        if len(self.__fields__) < 2:
            return super().__repr__()
        lines = [f"{self.__class__.__name__}("]
        for name in self.__fields__.keys():
            field_val = getattr(self, name)
            if name in redacted_fields:
                lines.append(f"\t{name}=********,".replace("\n", "\n\t"))
            elif isinstance(field_val, list) and len(field_val) > 1:
                lines.append(f"\t{name}=[")
                lines.extend([f"\t\t{val!r},".replace("\n", "\n\t\t") for val in field_val])
                lines.append("\t],")
            elif isinstance(field_val, dict) and len(field_val) > 1:
                lines.append(f"\t{name}=" + "{")
                field_lines = [f"\t\t{key}={val!r}," for key, val in field_val.items()]
                lines.extend([line.replace("\n", "\n\t\t") for line in field_lines])
                lines.append("\t},")
            elif isinstance(field_val, Enum):
                lines.append(f"\t{name}={field_val.value!r},")
            else:
                lines.append(f"\t{name}={field_val!r},".replace("\n", "\n\t"))
        lines.append(")")
        return "\n".join(lines).replace("\t", "    ")


class NamedComponent(ComponentModel):
    """A base class for AutoTransform components that are produced from a factory using
    a name to differentiate component classes.

    Attributes:
        name (ClassVar[Enum]): The name of the Component.
    """

    name: ClassVar[Enum]

    def bundle(self) -> Dict[str, Any]:
        """Generates a JSON encodable bundle.
        If a component is not JSON encodable this method should be overridden to provide
        an encodable version.

        Returns:
            Dict[str, Any]: The encodable bundle.
        """

        return {"name": self.name} | super().bundle()


class ComponentImport(ComponentModel):
    """The information required to import and return a component.

    Attributes:
        class_name (str): The name of the class of the component.
        module (str): The fully qualified module where the class can be imported.
    """

    class_name: str
    module: str


T = TypeVar("T", bound=NamedComponent)


class ComponentFactory(Generic[T], ABC):
    """A factory class that produces instances of components

    Attributes:
        _components (Dict[str, ComponentImport]): A mapping from a component name
            to the information needed to import and use that component.
        _custom_components_file (str): The name of the file containing custom component JSON.
        _type (Type[T]): The type this factory will produce.
    """

    _components: Dict[str, ComponentImport]
    _custom_components_file: str
    _type: Type[T]

    def __init__(
        self, components: Dict[str, ComponentImport], component_type: Type[T], component_file: str
    ):
        """A simple constructor

        Args:
            components (Dict[str, ComponentImport]): The library of components.
            component_type (Type[T]): The type this factory produces.
            component_file (str): The name of the file containing custom component JSON.
        """
        self._components = components
        self._custom_components_file = component_file
        self._type = component_type

    def get_components(self) -> Dict[str, ComponentImport]:
        """A simple getter for the factory's map of components.

        Returns:
            Dict[str, ComponentImport]: The components in the component library.
        """
        return self._components

    def get_custom_components(self, strict: bool = False) -> Dict[str, ComponentImport]:
        """Builds the custom components dictionary if it's not set, returns it otherwise.

        Args:
            strict (bool, optional): Whether to strictly enforce no issues arise during
                import. Defaults to False

        Returns:
            Dict[str, ComponentImport]: The custom component import info.
        """

        return self._custom_components_strict if strict else self._custom_components

    @cached_property
    def _custom_components(self) -> Dict[str, ComponentImport]:
        """A cached property for the non-strict custom components.

        Returns:
            Dict[str, ComponentImport]: Non-strict custom components.
        """

        return ComponentFactory.get_custom_components_dict(
            self._custom_components_file, strict=False
        )

    @cached_property
    def _custom_components_strict(self) -> Dict[str, ComponentImport]:
        """A cached property for the strict custom components.

        Returns:
            Dict[str, ComponentImport]: Strict custom components.
        """

        return ComponentFactory.get_custom_components_dict(
            self._custom_components_file, strict=True
        )

    # pylint: disable=too-many-arguments,too-many-branches
    def from_console(
        self,
        name: str,
        previous_value: Optional[T] = UNSET_VALUE,  # type: ignore
        default_value: Optional[T] = UNSET_VALUE,  # type: ignore
        simple: bool = False,
        allow_none: bool = True,
    ) -> Optional[T]:
        """Gets a component from console inputs.

        Args:
            name (str): The name of the component being entered.
            previous_value (Optional[T], optional): A previously used value for the
                component. Defaults to UNSET_VALUE.
            default_value (Optional[T], optional): A default value for the component.
                Defaults to UNSET_VALUE.
            simple (bool, optional): Whether to choose simple options. Defaults to False.
            allow_none (bool, optional): Whether None is a valid value. Defaults to True.

        Returns:
            Optional[T]: The component or None.
        """

        if (isinstance(previous_value, self._type) or (previous_value is None and allow_none)) and (
            simple or choose_yes_or_no(f"Use previous {name}: {previous_value!r}?")
        ):
            return previous_value

        # pylint: disable=too-many-boolean-expressions
        if (
            (isinstance(default_value, self._type) or (default_value is None and allow_none))
            and default_value != previous_value
            and (simple or choose_yes_or_no(f"Use default {name}: {default_value!r}?"))
        ):
            return default_value
        all_components = self.get_components() | self.get_custom_components()
        options: List[Tuple[str, str]] = [(key, key) for key in all_components.keys()]
        component_name = choose_options_from_list(
            f"Choose a {name}", options, min_choices=0 if allow_none else 1
        )
        if allow_none and not bool(component_name):
            return None
        component_class = self.get_class(component_name[0])
        if not bool(component_class.__fields__):
            return component_class.from_data({})
        info(f"{component_class.__name__} Fields:")
        for field_name, field in component_class.__fields__.items():
            # pylint: disable=protected-access
            if field.required:
                info(f"\t{field_name}: {field._type_display()}")
            else:
                info(f"\t{field_name}: {field._type_display()} = {field.get_default()}")

        while True:
            component_json = get_str(f"Enter JSON encoded {component_class.__name__}: ")
            try:
                if component_json != "":
                    component_data = json.loads(component_json)
                else:
                    component_data = {}
                if not isinstance(component_data, Dict):
                    error("Invalid JSON data, must be Dict")
                    continue
                return component_class.from_data(component_data)
            except json.JSONDecodeError as err:
                error(f"Failed to parse JSON\n{err}")
            except TypeError as err:
                error(f"Invalid type found\n{err}")
            except ValueError as err:
                error(f"Invalid value found\n{err}")

    def get_instance(self, data: Dict[str, Any]) -> T:
        """Simple method to get an instance from a bundle.

        Args:
            data (Dict[str, Any]): The bundled component.

        Returns:
            T: An instance of the associated component.
        """

        return self.get_class(data["name"]).from_data(data)

    def get_class(self, component_name: str) -> Type[T]:
        """Gets the class for a component with the specified type, usually an enum value.

        Args:
            component_name (str): The type of the component, usually an enum value. If a custom
                component is used, the type should start with custom/.

        Raises:
            ValueError: If the component could not be found.

        Returns:
            Type[T]: The class for the component.
        """

        component_info = self.get_components().get(component_name)
        if component_info is not None:
            return self._get_component_class(component_info)
        custom_components = self.get_custom_components()
        component_info = custom_components.get(component_name)
        if component_info is not None:
            return self._get_component_class(component_info)

        names = json.dumps(list(self.get_components().keys() | self.get_custom_components().keys()))
        raise ValueError(f"No component found with name {component_name}, valid names: {names}")

    @staticmethod
    def get_custom_components_dict(
        component_file_name: str, strict: bool = False
    ) -> Dict[str, ComponentImport]:
        """Builds a custom component dictionary for importing.

        Args:
            component_file_name (str): The name of the file that contains the custom components.
            strict (bool, optional): Whether to raise an error if JSON is malformed.
                Defaults to False.

        Returns:
            Dict[str, ComponentImport]: The custom component import info. All custom components
                start with custom/ for their name.
        """

        custom_components: Dict[str, ComponentImport] = {}
        component_json_path = ComponentFactory.get_custom_components_path(component_file_name)
        EventHandler.get().handle(
            DebugEvent({"message": f"Importing custom components from: {component_json_path}"})
        )
        try:
            with open(component_json_path, "r", encoding="UTF-8") as component_file:
                json_components = json.load(component_file)
        except FileNotFoundError:
            EventHandler.get().handle(DebugEvent({"message": "No components file."}))
            json_components = {}
        if not isinstance(json_components, Dict):
            message = f"Malformed custom component file: {component_json_path}"
            if strict:
                raise ValueError(message)
            EventHandler.get().handle(WarningEvent({"message": message}))
            return custom_components

        for name, import_info in json_components.items():
            if not isinstance(name, str):
                message = f"Invalid name: {name}"
                if strict:
                    raise ValueError(message)
                EventHandler.get().handle(WarningEvent({"message": message}))
                continue
            try:
                custom_components[f"custom/{name}"] = ComponentImport.from_data(import_info)
            except Exception as err:  # pylint: disable=broad-except
                if strict:
                    raise err
                EventHandler.get().handle(WarningEvent({"message": str(err)}))
        return custom_components

    @staticmethod
    def get_custom_components_path(component_file_name: str) -> str:
        """Gets the path for the custom component file.

        Args:
            component_file_name (str): The name of the file that contains the custom components.

        Returns:
            str: The path where the custom component JSON is located.
        """

        component_directory = get_config().component_directory
        if component_directory is None:
            component_directory = "autotransform"

        return f"{component_directory}/{component_file_name}"

    def _get_component_class(
        self,
        component_info: ComponentImport,
    ) -> Type[T]:
        """Gets the class from the ComponentImport info.

        Args:
            component_info (ComponentImport): The info required to import the component.

        Returns:
            Type[T]: The class for the component.
        """

        module = importlib.import_module(component_info.module)
        assert hasattr(module, component_info.class_name)
        component = getattr(module, component_info.class_name)
        assert isinstance(component, type), "Component is not a class"
        assert issubclass(
            component, self._type
        ), f"Component must be a subclass of {str(self._type)}"
        return component



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Provides utility methods for interacting with the Schema map."""

from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Dict, Optional, Sequence, Tuple

from autotransform.config import get_repo_config_relative_path
from autotransform.schema.builder import FACTORY as schema_builder_factory
from autotransform.schema.schema import AutoTransformSchema
from autotransform.util.enums import SchemaType


class SchemaMap:
    """A map that can convert Schema names in to AutoTransformSchemas.

    Attributes:
        _data (Dict[str, Tuple[SchemaType, str]]): The data that backs the Schema map.
        __instance (Optional[SchemaMap]): The singleton instance of the SchemaMap.
    """

    _data: Dict[str, Tuple[SchemaType, str]]

    __instance: Optional[SchemaMap] = None

    def __init__(self):
        """A simple constructor for a singleton class"""

        assert SchemaMap.__instance is None

        map_file_path = SchemaMap.get_schema_map_path()
        if Path(map_file_path).is_file():
            with open(map_file_path, "r", encoding="UTF-8") as map_file:
                schema_map = json.loads(map_file.read())
        else:
            schema_map = {}

        self._data = {
            name: (SchemaType(schema["type"]), schema["target"])
            for name, schema in schema_map.items()
        }

    @staticmethod
    def get() -> SchemaMap:
        """Gets the singleton instance of the SchemaMap.

        Returns:
            SchemaMap: The singleton instance of the SchemaMap.
        """

        if SchemaMap.__instance is None:
            SchemaMap.__instance = SchemaMap()

        return SchemaMap.__instance

    @staticmethod
    def get_schema_directory() -> str:
        """Gets the directory where schemas and the schema map are located.

        Returns:
            str: The directory where schemas and the schema map are located.
        """

        return os.getenv(
            "AUTO_TRANSFORM_SCHEMA_DIRECTORY",
            f"{get_repo_config_relative_path()}/schemas",
        )

    @staticmethod
    def get_schema_map_path() -> str:
        """Gets the path to the schema map.

        Returns:
            str: The path to the schema map.
        """

        return f"{SchemaMap.get_schema_directory()}/schema_map.json"

    @staticmethod
    def get_schema_path(file_name: str) -> str:
        """Gets the path to a given schema

        Args:
            file_name (str): The name of the file for a schema.

        Returns:
            str: The path where a schema is located.
        """

        return f"{SchemaMap.get_schema_directory()}/{file_name}"

    def get_schema(self, schema_name: str) -> AutoTransformSchema:
        """Get an AutoTransformSchema from the map.

        Args:
            schema_name (str): The name of the schema to get.

        Returns:
            AutoTransformSchema: The AutoTransformSchema with the supplied name.
        """

        schema_type, target = self._data[schema_name]

        if schema_type == SchemaType.BUILDER:
            return schema_builder_factory.get_instance({"name": target}).build()

        with open(SchemaMap.get_schema_path(target), "r", encoding="UTF-8") as schema_file:
            return AutoTransformSchema.from_data(json.loads(schema_file.read()))

    def add_schema(self, schema_name: str, schema_type: SchemaType, target: str) -> None:
        """Adds the specified Schema to the map.

        Args:
            schema_name (str): The name of the Schema to add.
            schema_type (SchemaType): The type of the Schema.
            target (str): The target for the Schema.
        """

        self._data[schema_name] = (schema_type, target)

    def remove_schema(self, schema_name: str) -> None:
        """Removes the specified Schema from the map.

        Args:
            schema_name (str): The name of the Schema to remove.
        """

        del self._data[schema_name]

    def write(self) -> None:
        """Writes the SchemaMap to a file as JSON."""

        schema_map_data = {
            name: {"type": schema[0], "target": schema[1]} for name, schema in self._data.items()
        }

        file_path = SchemaMap.get_schema_map_path()
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w+", encoding="UTF-8") as schema_map_file:
            schema_map_file.write(json.dumps(schema_map_data, indent=4))
            schema_map_file.flush()

    def items(self) -> Sequence[Tuple[str, Tuple[SchemaType, str]]]:
        """Gets the items in the SchemaMap.

        Returns:
            Tuple[str, Tuple[SchemaType, str]]: The items from the SchemaMap.
        """

        return list(self._data.items())

    def __contains__(self, schema_name: str) -> bool:
        """Override the contains method for the map.

        Args:
            schema_name (str): The name to check for in the SchemaMap.

        Returns:
            bool: Whether a Schema with the given name is in the map.
        """

        return schema_name in self._data

    def __getitem__(self, schema_name: str) -> AutoTransformSchema:
        """Override the subscript operator.

        Args:
            schema_name (str): The name of the Schema to get.

        Returns:
            AutoTransformSchema: The Schema with the specified name.
        """

        return self.get_schema(schema_name)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for script based Filters."""

from __future__ import annotations

import json
import subprocess
from tempfile import NamedTemporaryFile as TmpFile
from typing import ClassVar, List, Optional, Sequence, Set

from autotransform.event.handler import EventHandler
from autotransform.event.verbose import VerboseEvent
from autotransform.filter.base import BulkFilter, FilterName
from autotransform.item.base import Item
from autotransform.util.functions import replace_script_args


class ScriptFilter(BulkFilter):
    """A Filter that uses a script to validate Items.

    Attributes:
        args (List[str]): The arguments to supply to the script.
        script (str): The script to run.
        timeout (int): The timeout to use for the script process.
        chunk_size (Optional[int], optional): The maximum number of items per run of the
            script. If None, then no chunking is used. Defaults to None.
        name (ClassVar[FilterName]): The name of the component.
    """

    args: List[str]
    script: str
    timeout: int

    chunk_size: Optional[int] = None

    name: ClassVar[FilterName] = FilterName.SCRIPT

    def _get_valid_keys(self, items: Sequence[Item]) -> Set[str]:
        """Gets the valid keys from the Items using a script. If a <<RESULT_FILE>> arg is used
        it will be replaced with the path of a temporary file that can be used to store a JSON
        encoded list of keys for valid Items. If no such arg is used, the STDOUT of the script
        will be interpreted as a JSON encoded list of keys for valid Items. Additionally, the
        <<ITEM_FILE>> argument will be replaced with the path to a file containing a JSON
        encoded list of the items to validate.

        Args:
            items (Sequence[Item]): The Items to check for valid items.

        Returns:
            Set[str]: The keys of the valid Items.
        """

        event_handler = EventHandler.get()

        # Get Command
        cmd = [self.script]
        cmd.extend(self.args)

        chunk_size = self.chunk_size or len(items)
        item_chunks = [items[i : i + chunk_size] for i in range(0, len(items), chunk_size)]

        valid_keys: Set[str] = set()
        for chunk in item_chunks:
            with TmpFile(mode="r+b") as res_file, TmpFile(mode="w+") as item_file:
                json.dump([item.bundle() for item in chunk], item_file)
                item_file.flush()
                arg_replacements = {
                    "<<RESULT_FILE>>": [res_file.name],
                    "<<ITEM_FILE>>": [item_file.name],
                }
                uses_result_file = "<<RESULT_FILE>>" in cmd
                replaced_cmd = replace_script_args(cmd, arg_replacements)

                # Run script
                event_handler.handle(VerboseEvent({"message": f"Running command: {replaced_cmd}"}))
                proc = subprocess.run(
                    replaced_cmd,
                    capture_output=True,
                    encoding="utf-8",
                    check=False,
                    timeout=self.timeout,
                )

                if proc.stdout.strip() != "" and uses_result_file:
                    event_handler.handle(
                        VerboseEvent({"message": f"STDOUT:\n{proc.stdout.strip()}"}),
                    )
                elif uses_result_file:
                    event_handler.handle(VerboseEvent({"message": "No STDOUT"}))

                if proc.stderr.strip() != "":
                    event_handler.handle(
                        VerboseEvent({"message": f"STDERR:\n{proc.stderr.strip()}"}),
                    )
                else:
                    event_handler.handle(VerboseEvent({"message": "No STDERR"}))
                proc.check_returncode()
                if uses_result_file:
                    with open(res_file.name, encoding="utf-8") as results:
                        key_data = json.loads(results.read())
                else:
                    key_data = json.loads(proc.stdout.strip())
                valid_keys = valid_keys.union(set(key_data))
        return valid_keys



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The base class and associated classes for Change components."""

from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, ClassVar, List

from autotransform.batcher.base import Batch
from autotransform.util.component import ComponentFactory, ComponentImport, NamedComponent

if TYPE_CHECKING:
    from autotransform.runner.base import Runner
    from autotransform.schema.schema import AutoTransformSchema


class ChangeState(str, Enum):
    """A simple enum for the state of a given Change in code review or version
    control systems."""

    CLOSED = "closed"
    MERGED = "merged"
    OPEN = "open"


class ReviewState(str, Enum):
    """A simple enum for the review state of a given Change in code review or version
    control systems."""

    APPROVED = "approved"
    CHANGES_REQUESTED = "changes_requested"
    NEEDS_REVIEW = "needs_review"


class TestState(str, Enum):
    """A simple enum for the test state of a given Change in code review or version
    control systems."""

    FAILURE = "failure"
    PENDING = "pending"
    SUCCESS = "success"


class ChangeName(str, Enum):
    """A simple enum for mapping."""

    GITHUB = "github"


class Change(NamedComponent):
    """The base for Change components. Used by AutoTransform to manage submissions to
    code review and source control systems.

    Attributes:
        name (ClassVar[ChangeName]): The name of the Component.
    """

    name: ClassVar[ChangeName]

    @abstractmethod
    def get_batch(self) -> Batch:
        """Gets the Batch that was used to produce the Change.

        Returns:
            Batch: The Batch used to produce the Change.
        """

    @abstractmethod
    def get_schema(self) -> AutoTransformSchema:
        """Gets the Schema that was used to produce the Change.

        Returns:
            AutoTransformSchema: The Schema used to produce the Change.
        """

    def get_schema_name(self) -> str:
        """Gets the name of the Schema that produced the Change.

        Returns:
            str: The name of the Schema.
        """

        return self.get_schema().config.schema_name

    @abstractmethod
    def get_state(self) -> ChangeState:
        """Gets the current state of the Change.

        Returns:
            ChangeState: The current state of the Change.
        """

    @abstractmethod
    def get_mergeable_state(self) -> str:
        """Gets the mergeable state of the Change.

        Returns:
            ChangeState: The mergeable state of the Change.
        """

    @abstractmethod
    def get_review_state(self) -> ReviewState:
        """Gets the current review state of the Change.

        Returns:
            ReviewState: The current review state of the Change.
        """

    @abstractmethod
    def get_test_state(self) -> TestState:
        """Gets the current test state of the Change.

        Returns:
            TestState: The current test state of the Change.
        """

    @abstractmethod
    def get_labels(self) -> List[str]:
        """Gets all labels for a Change.

        Returns:
            List[str]: The list of labels.
        """

    @abstractmethod
    def get_reviewers(self) -> List[str]:
        """Gets all reviewers for a Change.

        Returns:
            List[str]: The list of reviewers.
        """

    @abstractmethod
    def get_team_reviewers(self) -> List[str]:
        """Gets all team reviewers for a Change.

        Returns:
            List[str]: The list of team reviewers.
        """

    @abstractmethod
    def get_created_timestamp(self) -> int:
        """Returns the timestamp when the Change was created.

        Returns:
            int: The timestamp in seconds when the Change was created.
        """

    @abstractmethod
    def get_last_updated_timestamp(self) -> int:
        """Returns the timestamp when the Change was last updated.

        Returns:
            int: The timestamp in seconds when the Change was last updated.
        """

    @abstractmethod
    def abandon(self) -> bool:
        """Close out and abandon a Change, removing it from the code review
        and/or version control system.

        Returns:
            bool: Whether the abandon was completed successfully.
        """

    @abstractmethod
    def add_labels(self, labels: List[str]) -> bool:
        """Adds labels to an outstanding Change.

        Args:
            labels (List[str]): The labels to add.

        Returns:
            bool: Whether the labels were added successfully.
        """

    @abstractmethod
    def add_reviewers(self, reviewers: List[str], team_reviewers: List[str]) -> bool:
        """Adds reviewers to an outstanding Change.

        Args:
            reviewers (List[str]): The reviewers to add.
            team_reviewers (List[str]): Any team reviewers to add.

        Returns:
            bool: Whether the reviewers were added successfully.
        """

    @abstractmethod
    def comment(self, body: str) -> bool:
        """Comments on an outstanding Change.

        Args:
            body (str): The body of the comment.

        Returns:
            bool: Whether the comment was successful.
        """

    @abstractmethod
    def merge(self) -> bool:
        """Merges an approved change in to main.

        Returns:
            bool: Whether the merge was completed successfully.
        """

    @abstractmethod
    def remove_label(self, label: str) -> bool:
        """Removes a label from an outstanding Change.

        Args:
            label (str): The label to remove.

        Returns:
            bool: Whether the label was removed successfully.
        """

    def update(self, runner: Runner) -> bool:
        """Update an outstanding Change against the latest state of the codebase.

        Args:
            runner (Runner): The Runner to use to update the Change.

        Returns:
            bool: Whether the update was completed successfully.
        """

        runner.update(self)
        return True


FACTORY = ComponentFactory(
    {
        ChangeName.GITHUB: ComponentImport(
            class_name="GithubChange", module="autotransform.change.github"
        ),
    },
    Change,  # type: ignore [type-abstract]
    "change.json",
)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for inline Inputs."""

from __future__ import annotations

from typing import Any, ClassVar, Dict, List, Sequence, Type

from autotransform.input.base import Input, InputName
from autotransform.item.base import FACTORY as item_factory
from autotransform.item.base import Item
from autotransform.item.file import FileItem


class InlineInput(Input):
    """An Input that simply returns the supplied Items as input.

    Attributes:
        items (List[Item]): The items to return.
        name (ClassVar[InputName]): The name of the component.
    """

    items: List[Item]

    name: ClassVar[InputName] = InputName.INLINE

    def get_items(self) -> Sequence[Item]:
        """Returns the supplied Items as input.

        Returns:
            Sequence[Item]: The supplied Items.
        """

        return self.items

    @classmethod
    def from_data(cls: Type[InlineInput], data: Dict[str, Any]) -> InlineInput:
        """Produces an instance of the component from decoded data.

        Args:
            data (Dict[str, Any]): The JSON decoded data.

        Returns:
            InlineInput: An instance of the component.
        """

        return cls(items=[item_factory.get_instance(item) for item in data["items"]])


class InlineFileInput(Input):
    """An Input that simply returns the supplied file paths as FileItems for input.

    Attributes:
        files (List[str]): The file paths for Items to return.
        name (ClassVar[InputName]): The name of the component.
    """

    files: List[str]

    name: ClassVar[InputName] = InputName.INLINE_FILE

    def get_items(self) -> Sequence[FileItem]:
        """Returns the supplied files as FileItems for input.

        Returns:
            Sequence[FileItem]: The supplied files as FileItems.
        """

        return [FileItem(key=file) for file in self.files]


class InlineGenericInput(Input):
    """An Input that simply returns the supplied keys as Items for input.

    Attributes:
        keys (List[str]): The keys for Items to return.
        name (ClassVar[InputName]): The name of the component.
    """

    keys: List[str]

    name: ClassVar[InputName] = InputName.INLINE_GENERIC

    def get_items(self) -> Sequence[Item]:
        """Returns the supplied keys as Items for input.

        Returns:
            Sequence[Item]: The supplied keys as Items.
        """

        return [Item(key=key) for key in self.keys]



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The schedule command is used to schedule runs of AutoTransform."""

import time
from argparse import ArgumentParser, Namespace

from autotransform.config import get_config, get_repo_config_relative_path
from autotransform.event.handler import EventHandler
from autotransform.event.logginglevel import LoggingLevel
from autotransform.event.run import ScriptRunEvent
from autotransform.event.verbose import VerboseEvent
from autotransform.runner.local import LocalRunner
from autotransform.util.scheduler import Scheduler


def add_args(parser: ArgumentParser) -> None:
    """Adds the args to a subparser that are required to schedule runs.

    Args:
        parser (ArgumentParser): The parser for the schedule command.
    """

    parser.add_argument(
        "--path",
        metavar="path",
        required=False,
        type=str,
        help="A file path to the JSON encoded file, only use if file is in a non-usual place.",
    )

    parser.add_argument(
        "-t",
        "--time",
        metavar="time",
        type=int,
        required=False,
        help="The timestamp to use in place of the current time, used in cases where delays in "
        + "scheduling are likely.",
    )

    logging_level = parser.add_mutually_exclusive_group()
    logging_level.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        required=False,
        help="Tells the script to output verbose logs.",
    )
    logging_level.add_argument(
        "-d",
        "--debug",
        dest="debug",
        action="store_true",
        required=False,
        help="Tells the script to output debug logs.",
    )

    mode_group = parser.add_mutually_exclusive_group()
    mode_group.add_argument(
        "-l",
        "--local",
        dest="run_local",
        action="store_true",
        required=False,
        help="Tells the script to use the local runner.",
    )
    mode_group.add_argument(
        "-r",
        "--remote",
        dest="run_local",
        action="store_false",
        required=False,
        help="Tells the script to use the remote runner. This is the default mode.",
    )

    parser.set_defaults(run_local=False, func=schedule_command_main)


def schedule_command_main(args: Namespace) -> None:
    """The main method for the schedule command, handles the actual execution of scheduling runs.

    Args:
        args (Namespace): The arguments supplied to the schedule command, such as the JSON file.
    """

    # pylint: disable=unspecified-encoding

    start_time = int(args.time) if args.time is not None else int(time.time())
    event_handler = EventHandler.get()
    if args.verbose:
        event_handler.set_logging_level(LoggingLevel.VERBOSE)
    if args.debug:
        event_handler.set_logging_level(LoggingLevel.DEBUG)

    # Get Scheduler
    schedule_file = args.path
    if schedule_file is None:
        schedule_file = f"{get_repo_config_relative_path()}/scheduler.json"
    event_args = {"scheduler_file": schedule_file}
    scheduler = Scheduler.read(schedule_file)
    event_args["scheduler"] = scheduler
    event_handler.handle(ScriptRunEvent({"script": "schedule", "args": event_args}))

    event_handler.get().handle(VerboseEvent({"message": f"Running scheduler: {scheduler!r}"}))
    if args.run_local:
        runner = get_config().local_runner
    else:
        runner = get_config().remote_runner
    scheduler.run(start_time, runner or LocalRunner())



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Filters are used by AutoTransform to take Items provided by Inputs and check them for
eligibility. Items that pass Filters will get passed along to Batchers."""



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the LibCSTTransformer. See https://github.com/Instagram/LibCST"""

from __future__ import annotations

import importlib
from functools import cached_property
from typing import Any, ClassVar, Dict

from autotransform.event.debug import DebugEvent
from autotransform.event.handler import EventHandler
from autotransform.event.warning import WarningEvent
from autotransform.item.base import Item
from autotransform.item.file import FileItem
from autotransform.transformer.base import TransformerName
from autotransform.transformer.single import SingleTransformer
from libcst.codemod import (
    CodemodCommand,
    CodemodContext,
    TransformExit,
    TransformFailure,
    TransformSkip,
    TransformSuccess,
    transform_module,
)
from pydantic import Field


class LibCSTTransformer(SingleTransformer):
    """A Transformer that makes changes using LibCST.

    Attributes:
        command_module (str): The fully qualified module where the CodemodCommand is located.
        command_name (str): The name of the class for the CodemodCommand.
        command_args (optional, Dict[str, Any]): Arguments for the CodemodCommand. Defaults to {}.
        name (ClassVar[TransformerName]): The name of the component.
    """

    command_module: str
    command_name: str

    command_args: Dict[str, Any] = Field(default_factory=dict)

    name: ClassVar[TransformerName] = TransformerName.LIBCST

    def _transform_item(self, item: Item) -> None:
        """Run the supplied CodemodCommand against a FileItem.

        Args:
            item (Item): The file that will be transformed.
        """

        assert isinstance(item, FileItem)
        event_handler = EventHandler.get()
        event_handler.handle(DebugEvent({"message": f"Performing transform on {item.get_path()}"}))
        res = transform_module(self._command, item.get_content())
        for message in res.warning_messages:
            event_handler.handle(WarningEvent({"message": f"Warning: {message}"}))
        if isinstance(res, TransformSuccess):
            event_handler.handle(DebugEvent({"message": "Transform success"}))
            item.write_content(res.code)
        if isinstance(res, TransformSkip):
            event_handler.handle(
                DebugEvent(
                    {"message": f"Transform skipped ({res.skip_reason}): {res.skip_description}"}
                )
            )
        if isinstance(res, TransformFailure):
            event_handler.handle(
                WarningEvent({"message": f"Transform failed: {res.error}\n{res.traceback_str}"})
            )
        if isinstance(res, TransformExit):
            event_handler.handle(DebugEvent({"message": "Transform exited from user interupt"}))

    @cached_property
    def _command(self) -> CodemodCommand:
        """Gets an instance of the command and caches it.

        Returns:
            CodemodCommand: An instance of the command to run.
        """

        module = importlib.import_module(self.command_module)
        assert hasattr(module, self.command_name)
        command_class = getattr(module, self.command_name)
        assert isinstance(command_class, type), "Command is not a class"
        assert issubclass(
            command_class, CodemodCommand
        ), "Component must be a subclass of CodemodCommand"
        return command_class(CodemodContext(), **self.command_args)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the ConditionalStep."""

from __future__ import annotations

from typing import Any, ClassVar, Dict, List, Type

from autotransform.change.base import Change
from autotransform.step.action.base import FACTORY as action_factory
from autotransform.step.action.base import Action, ActionName
from autotransform.step.base import Step, StepName
from autotransform.step.condition.base import FACTORY as condition_factory
from autotransform.step.condition.base import Condition


class ConditionalStep(Step):
    """A step which takes action based on whether a particular condition is passed. Allows
    for creating customized steps based on different Change attributes readily through JSON.

    Attributes:
        actions (List[Action]): The Actions to perform if the Condition passes.
        condition (Condition): The condition to check.
        continue_if_passed (bool, optional): Whether to continue on to the next step after
            performing the action if the condition passes. Defaults to False.
        name (ClassVar[StepName]): The name of the component.
    """

    actions: List[Action]
    condition: Condition
    continue_if_passed: bool = False

    name: ClassVar[StepName] = StepName.CONDITIONAL

    def get_actions(self, change: Change) -> List[Action]:
        """Checks the Change against the provided Condition and returns the appropriate Actions
        based on whether or not the Condition is passed. If no Actions are returned, the Step
        is skipped.

        Args:
            change (Change): The Change the Step is running against.

        Returns:
            List[Action]: The Actions the Step wants to take.
        """

        if self.condition.check(change):
            return self.actions
        return []

    def continue_management(self, change: Change) -> bool:
        """Checks if management should be continued after this Step when Actions were provided.

        Args:
            change (Change): The Change the Step is running against.

        Returns:
            bool: Whether to continue management.
        """
        return self.continue_if_passed

    @classmethod
    def from_data(cls: Type[ConditionalStep], data: Dict[str, Any]) -> ConditionalStep:
        """Produces an instance of the component from decoded data.

        Args:
            data (Mapping[str, Any]): The JSON decoded data.

        Returns:
            ConditionalStep: An instance of the component.
        """

        if "action" in data:
            action_name = (
                data["action"]
                if isinstance(data["action"], ActionName)
                else ActionName(data["action"])
            )
            actions = [action_factory.get_instance({"name": action_name})]
        else:
            actions = [action_factory.get_instance(action) for action in data["actions"]]
        condition = condition_factory.get_instance(data["condition"])
        if "continue_if_passed" in data:
            continue_if_passed = bool(data["continue_if_passed"])
        else:
            continue_if_passed = False
        return cls(actions=actions, condition=condition, continue_if_passed=continue_if_passed)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The heart of AutoTransform, AutoTransformSchemas represent all information
required to fully execute a change."""

from __future__ import annotations

import json
from typing import Any, Dict, List, Optional

import autotransform.schema
from autotransform.batcher.base import FACTORY as batcher_factory
from autotransform.batcher.base import Batch, Batcher
from autotransform.change.base import Change
from autotransform.command.base import FACTORY as command_factory
from autotransform.command.base import Command
from autotransform.event.debug import DebugEvent
from autotransform.event.handler import EventHandler
from autotransform.event.verbose import VerboseEvent
from autotransform.filter.base import FACTORY as filter_factory
from autotransform.filter.base import BulkFilter, Filter
from autotransform.input.base import FACTORY as input_factory
from autotransform.input.base import Input
from autotransform.item.base import Item
from autotransform.repo.base import FACTORY as repo_factory
from autotransform.repo.base import Repo
from autotransform.schema.config import SchemaConfig
from autotransform.transformer.base import FACTORY as transformer_factory
from autotransform.transformer.base import Transformer
from autotransform.util.component import ComponentModel
from autotransform.util.console import choose_options_from_list, choose_yes_or_no
from autotransform.validator.base import FACTORY as validator_factory
from autotransform.validator.base import ValidationError, Validator
from pydantic import Field


class AutoTransformSchema(ComponentModel):
    """The heart of AutoTransform, pulls together all components required to execute
    a transformation.

    Attributes:
        input (Input): The Input which gets Items.
        batcher (Batcher): The Batcher which batches filtered Items in to logical groups.
        transformer (Transformer): The Transformer which actually modifies files.
        config (SchemaConfig): Any configuration needed by the schema so that it can run.
        filters (List[Filter], optional): A list of Filters to apply to Items. Defaults to [].
        validators (List[Validator], optional): A list of Validators to ensure the changes
            did not break anything. Defaults to [].
        commands (List[Command], optional): A list of Commands that run post-processing on
            the changes. Defaults to [].
        repo (Optional[Repo], optional): A Repo to control submission of changes to version
            control or code review systems. Defaults to None.
    """

    # pylint: disable=too-many-instance-attributes

    input: Input
    batcher: Batcher
    transformer: Transformer
    config: SchemaConfig

    filters: List[Filter] = Field(default_factory=list)
    validators: List[Validator] = Field(default_factory=list)
    commands: List[Command] = Field(default_factory=list)
    repo: Optional[Repo] = None

    def get_batches(self) -> List[Batch]:
        """Runs the Input to get eligible Items, filters them, then batches them.
        Note: this function is not thread safe.

        Returns:
            List[Batch]: The Batches for the change
        """

        autotransform.schema.current = self
        event_handler = EventHandler.get()
        event_handler.handle(VerboseEvent({"message": "Begin get_batches"}))

        # Get Items
        event_handler.handle(VerboseEvent({"message": "Begin get_items"}))
        all_items = self.input.get_items()
        item_str = "\n".join([f"{item!r}," for item in all_items])
        event_handler.handle(VerboseEvent({"message": f"Num Items: {len(all_items)}"}))
        event_handler.handle(DebugEvent({"message": f"Items: [\n{item_str}\n]"}))

        # Filter Items
        event_handler.handle(VerboseEvent({"message": "Begin filters"}))
        valid_items: List[Item] = []
        for filt in self.filters:
            if isinstance(filt, BulkFilter):
                filt.pre_process(all_items)
        for item in all_items:
            is_valid = True
            for cur_filter in self.filters:
                if not cur_filter.is_valid(item):
                    is_valid = False
                    event = DebugEvent({"message": f"[{cur_filter}] Invalid Item: {item!r}"})
                    event_handler.handle(event)
                    break
            if is_valid:
                valid_items.append(item)
        if valid_items:
            valid_item_str = "\n".join([f"{item!r}," for item in valid_items])
            event_handler.handle(VerboseEvent({"message": f"Num Valid Items: {len(valid_items)}"}))
            event_handler.handle(DebugEvent({"message": f"Valid items: [\n{valid_item_str}\n]"}))
        else:
            event_handler.handle(VerboseEvent({"message": "No valid items."}))

        # Batch Items
        event_handler.handle(VerboseEvent({"message": "Begin batching"}))
        batches = self.batcher.batch(valid_items)
        encodable_batches = [
            {"items": [item.bundle() for item in batch["items"]], "metadata": batch["metadata"]}
            for batch in batches
        ]
        event_handler.handle(VerboseEvent({"message": f"Num Batches: {len(batches)}"}))
        event_handler.handle(DebugEvent({"message": f"Batches: {json.dumps(encodable_batches)}"}))
        autotransform.schema.current = None

        return batches

    def execute_batch(self, batch: Batch, change: Optional[Change] = None) -> bool:
        """Executes changes for a batch, including setting up the Repo, running the Transformer,
        checking all Validators, running Commands, submitting changes if present, and rewinding
        the Repo if changes are submitted. Note: this function is not thread safe.

        Args:
            batch (Batch): The Batch to execute.
            change (Optional[Change]): An associated Change that is being updated.

        Raises:
            ValidationError: If one of the Schema's Validators fails raises an exception.

        Returns:
            bool: Whether the batch triggered a submission.
        """

        autotransform.schema.current = self
        event_handler = EventHandler.get()
        encodable_batch = {
            "items": [item.bundle() for item in batch["items"]],
            "metadata": batch["metadata"],
        }
        event_handler.handle(
            VerboseEvent(
                {"message": f"Handling Batch: {batch['title']} with {len(batch['items'])} items"}
            )
        )
        event_handler.handle(DebugEvent({"message": f"Full Batch: {json.dumps(encodable_batch)}"}))

        # Make sure repo is clean before executing
        if self.repo is not None:
            event_handler.handle(VerboseEvent({"message": "Clean repo"}))
            self.repo.rewind(batch)
            if change is None and self.repo.has_outstanding_change(batch):
                event_handler.handle(
                    VerboseEvent({"message": "Skipping batch with outstanding change"})
                )
                return False

        # Execute transformation
        result = self.transformer.transform(batch)

        # Run pre-validation commands
        pre_validation_commands = [
            command for command in self.commands if command.run_pre_validation
        ]
        for command in pre_validation_commands:
            event_handler.handle(VerboseEvent({"message": f"Running command {command}"}))
            command.run(batch, result)

        # Validate the changes
        for validator in self.validators:
            validation_result = validator.check(batch, result)
            event_handler.handle(
                VerboseEvent({"message": f"Validation Result: {validation_result}"})
            )

            if validation_result.level > self.config.allowed_validation_level:
                event_handler.handle(VerboseEvent({"message": "Validation Failed"}))
                raise ValidationError(issue=validation_result, message=validation_result.message)

        # Run post-validation commands
        post_validation_commands = [
            command for command in self.commands if not command.run_pre_validation
        ]
        for command in post_validation_commands:
            event_handler.handle(VerboseEvent({"message": f"Running command {command}"}))
            command.run(batch, result)

        submitted = False
        # Handle repo state, submitting changes if present and reseting the repo
        if self.repo is not None:
            event_handler.handle(VerboseEvent({"message": "Checking for changes"}))
            if self.repo.has_changes(batch):
                event_handler.handle(VerboseEvent({"message": "Changes found"}))
                event_handler.handle(VerboseEvent({"message": "Submitting changes"}))
                self.repo.submit(batch, result, change=change)
                event_handler.handle(VerboseEvent({"message": "Rewinding repo"}))
                self.repo.rewind(batch)
                submitted = True
            else:
                if change is not None:
                    event_handler.handle(
                        VerboseEvent({"message": "No changes in update, abandoning"})
                    )

                    change.abandon()
                event_handler.handle(VerboseEvent({"message": "No changes found"}))
        event_handler.handle(VerboseEvent({"message": "Finish batch"}))
        autotransform.schema.current = None
        return submitted

    def run(self):
        """Fully run a given Schema including getting and executing all Batches.
        Note: this function is not thread safe."""

        autotransform.schema.current = self
        batches = self.get_batches()
        num_submissions = 0
        for batch in batches:
            if (
                self.config.max_submissions is not None
                and num_submissions >= self.config.max_submissions
            ):
                EventHandler.get().handle(
                    VerboseEvent({"message": f"Max submissions reached: {num_submissions}"})
                )
                break
            if self.execute_batch(batch):
                num_submissions += 1
        autotransform.schema.current = None

    @staticmethod
    def from_data(data: Dict[str, Any]) -> AutoTransformSchema:
        """Takes data from a source like JSON and produces the associated Schema.

        Args:
            data (Dict[str, Any]): The data representing the Schema.

        Returns:
            AutoTransformSchema: The Schema represented by the data.
        """

        inp = input_factory.get_instance(data["input"])
        batcher = batcher_factory.get_instance(data["batcher"])
        transformer = transformer_factory.get_instance(data["transformer"])
        config = SchemaConfig.parse_obj(data["config"])

        filters = [filter_factory.get_instance(f) for f in data.get("filters", [])]
        validators = [
            validator_factory.get_instance(validator) for validator in data.get("validators", [])
        ]
        commands = [command_factory.get_instance(command) for command in data.get("commands", [])]

        repo = repo_factory.get_instance(data["repo"]) if "repo" in data else None

        return AutoTransformSchema(
            input=inp,
            batcher=batcher,
            transformer=transformer,
            filters=filters,
            validators=validators,
            commands=commands,
            repo=repo,
            config=config,
        )

    # pylint: disable=too-many-branches
    @staticmethod
    def from_console(prev_schema: Optional[AutoTransformSchema] = None) -> AutoTransformSchema:
        """Gets a AutoTransformSchema using console inputs.

        Args:
            prev_schema (Optional[AutoTransformSchema], optional): A previously input
                AutoTransformSchema. Defaults to None.

        Returns:
            AutoTransformSchema: The input AutoTransformSchema.
        """

        # Get Config
        config = SchemaConfig.from_console(prev_schema.config if prev_schema is not None else None)

        # Get Input
        args: Dict[str, Any] = {"allow_none": False}
        if prev_schema is not None:
            args["previous_value"] = prev_schema.input
        inp = input_factory.from_console("input", **args)
        assert inp is not None

        # Get Filters
        if prev_schema is not None and prev_schema.filters:
            filters = choose_options_from_list(
                "Choose filters to keep",
                [(filt, f"{filt!r}") for filt in prev_schema.filters],
                min_choices=0,
                max_choices=len(prev_schema.filters),
            )
        else:
            filters = []
        while choose_yes_or_no("Would you like to add a filter?"):
            filt = filter_factory.from_console("filter", allow_none=False)
            assert filt is not None
            filters.append(filt)

        # Get Batcher
        args = {"allow_none": False}
        if prev_schema is not None:
            args["previous_value"] = prev_schema.batcher
        batcher = batcher_factory.from_console("batcher", **args)
        assert batcher is not None

        # Get Transformer
        args = {"allow_none": False}
        if prev_schema is not None:
            args["previous_value"] = prev_schema.transformer
        transformer = transformer_factory.from_console("transformer", **args)
        assert transformer is not None

        # Get Validators
        if prev_schema is not None and prev_schema.validators:
            validators = choose_options_from_list(
                "Choose validators to keep",
                [(validator, f"{validator!r}") for validator in prev_schema.validators],
                min_choices=0,
                max_choices=len(prev_schema.validators),
            )
        else:
            validators = []
        while choose_yes_or_no("Would you like to add a validator?"):
            validator = validator_factory.from_console("validator", allow_none=False)
            assert validator is not None
            validators.append(validator)

        # Get Commands
        if prev_schema is not None and prev_schema.commands:
            commands = choose_options_from_list(
                "Choose commands to keep",
                [(command, f"{command!r}") for command in prev_schema.commands],
                min_choices=0,
                max_choices=len(prev_schema.commands),
            )
        else:
            commands = []
        while choose_yes_or_no("Would you like to add a command?"):
            command = command_factory.from_console("command", allow_none=False)
            assert command is not None
            commands.append(command)

        # Get Repo
        args = {"allow_none": False}
        if prev_schema is not None:
            args["previous_value"] = prev_schema.repo
        repo = repo_factory.from_console("repo", **args)
        assert repo is not None

        return AutoTransformSchema(
            input=inp,
            batcher=batcher,
            transformer=transformer,
            config=config,
            filters=filters,
            validators=validators,
            commands=commands,
            repo=repo,
        )



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the CodeownersBatcher."""

from __future__ import annotations

import math
from copy import deepcopy
from typing import Any, ClassVar, Dict, List, Optional, Sequence

from autotransform.batcher.base import Batch, Batcher, BatcherName
from autotransform.item.base import Item
from autotransform.item.file import FileItem
from codeowners import CodeOwners


class CodeownersBatcher(Batcher):
    """A batcher which uses Github CODEOWNERS files to separate changes by owner. Titles will
    be of the form 'prefix <owner>'

    Attributes:
        codeowners_location (str): The location of the CODEOWNERS file.
        prefix (str): The prefix to use for titles.
        max_batch_size (Optional[int]): The maximum size of any batch. If None, then batches will
            have no max size. Defaults to None.
        metadata (Optional[Dict[str, Any]], optional): The metadata to associate with
            Batches. Defaults to None.
        name (ClassVar[BatcherName]): The name of the Component.
    """

    codeowners_location: str
    prefix: str
    max_batch_size: Optional[int] = None
    metadata: Optional[Dict[str, Any]] = None

    name: ClassVar[BatcherName] = BatcherName.CODEOWNERS

    # pylint: disable=too-many-branches
    def batch(self, items: Sequence[Item]) -> List[Batch]:
        """Take filtered Items and batch them based on CODEOWNERS.

        Args:
            items (Sequence[Item]): The filtered Items to separate.

        Returns:
            List[Batch]: A list of Batches representing all items owned by a given
                owner.
        """

        team_owners: Dict[str, List[Item]] = {}
        individual_owners: Dict[str, List[Item]] = {}
        no_owners: List[Item] = []

        with open(self.codeowners_location, mode="r", encoding="UTF-8") as codeowners_file:
            codeowners = CodeOwners(codeowners_file.read())

        # Build Owner Dictionaries
        for item in items:
            assert isinstance(item, FileItem)
            owner = codeowners.of(item.get_path())

            if not owner:
                no_owners.append(item)
                continue

            owner_tuple = owner[0]
            if owner_tuple[0] == "USERNAME":
                owner_name = owner_tuple[1].removeprefix("@")
                if owner_name not in individual_owners:
                    individual_owners[owner_name] = []
                individual_owners[owner_name].append(item)

            if owner_tuple[0] == "TEAM":
                owner_name = owner_tuple[1].removeprefix("@")
                if owner_name not in team_owners:
                    team_owners[owner_name] = []
                team_owners[owner_name].append(item)

        batches: List[Batch] = []

        # Add batches based on team owners
        for team_owner, team_items in team_owners.items():
            if self.max_batch_size is not None and len(team_items) > self.max_batch_size:
                num_chunks = math.ceil(len(team_items) / self.max_batch_size)
                i = 1
                chunk_size = math.ceil(len(team_items) / num_chunks)
                item_chunks = [
                    team_items[i : i + chunk_size] for i in range(0, len(team_items), chunk_size)
                ]
                title = f"[{i}/{num_chunks}]{self.prefix} {team_owner}"
            else:
                num_chunks = 1
                i = 1
                item_chunks = [team_items]
                title = f"{self.prefix} {team_owner}"
            for chunk_items in item_chunks:
                batch: Batch = {"items": chunk_items, "title": title}
                # Deepcopy metadata to ensure mutations don't apply to all Batches
                metadata = deepcopy(self.metadata or {})
                if "team_reviewers" in metadata and team_owner not in metadata["team_reviewers"]:
                    metadata["team_reviewers"].append(team_owner)
                else:
                    metadata["team_reviewers"] = [team_owner]
                batch["metadata"] = metadata
                batches.append(batch)
                i += 1
                title = f"[{i}/{num_chunks}]{self.prefix} {team_owner}"

        # Add batches based on individual owners
        for individual_owner, individual_items in individual_owners.items():
            if self.max_batch_size is not None and len(individual_items) > self.max_batch_size:
                num_chunks = math.ceil(len(individual_items) / self.max_batch_size)
                i = 1
                chunk_size = math.ceil(len(individual_items) / num_chunks)
                item_chunks = [
                    individual_items[i : i + chunk_size]
                    for i in range(0, len(individual_items), chunk_size)
                ]
                title = f"[{i}/{num_chunks}]{self.prefix} {individual_owner}"
            else:
                num_chunks = 1
                i = 1
                item_chunks = [individual_items]
                title = f"{self.prefix} {individual_owner}"
            for chunk_items in item_chunks:
                batch = {"items": chunk_items, "title": title}
                # Deepcopy metadata to ensure mutations don't apply to all Batches
                metadata = deepcopy(self.metadata or {})
                if "reviewers" in metadata and individual_owner not in metadata["reviewers"]:
                    metadata["reviewers"].append(individual_owner)
                else:
                    metadata["reviewers"] = [individual_owner]
                batch["metadata"] = metadata
                batches.append(batch)
                i += 1
                title = f"[{i}/{num_chunks}]{self.prefix} {individual_owner}"

        # Add unowned batch
        if no_owners:
            if self.max_batch_size is not None and len(no_owners) > self.max_batch_size:
                num_chunks = math.ceil(len(no_owners) / self.max_batch_size)
                i = 1
                chunk_size = math.ceil(len(no_owners) / num_chunks)
                item_chunks = [
                    no_owners[i : i + chunk_size] for i in range(0, len(no_owners), chunk_size)
                ]
                title = f"[{i}/{num_chunks}]{self.prefix} unowned"
            else:
                num_chunks = 1
                i = 1
                item_chunks = [no_owners]
                title = f"{self.prefix} unowned"
            for chunk_items in item_chunks:
                batch = {"items": chunk_items, "title": title}
                if self.metadata is not None:
                    # Deepcopy metadata to ensure mutations don't apply to all Batches
                    batch["metadata"] = deepcopy(self.metadata)
                batches.append(batch)
                i += 1
                title = f"[{i}/{num_chunks}]{self.prefix} unowned"

            batches.append(batch)

        return batches



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The ManageActionEvent is triggered whenever an action is taken against an
outstanding Change.
"""

from typing import TypedDict

from autotransform.change.base import Change
from autotransform.event.base import Event
from autotransform.event.logginglevel import LoggingLevel
from autotransform.event.type import EventType
from autotransform.step.action.base import Action
from autotransform.step.base import Step


class ManageActionEventData(TypedDict):
    """The data for a ManageActionEventData. Contains the information that will be
    logged when the event is triggered."""

    action: Action
    change: Change
    step: Step


class ManageActionEvent(Event[ManageActionEventData]):
    """An event triggered whenever a Step attempts to take an action against a Change."""

    @staticmethod
    def get_type() -> EventType:
        """Used to represent the type of Event, output to logs.

        Returns:
            EventType: The unique type associated with this Event.
        """
        return EventType.MANAGE_ACTION

    @staticmethod
    def get_logging_level() -> LoggingLevel:
        """The logging level for events of this type.

        Returns:
            LoggingLevel: The logging detail required to log this event.
        """

        return LoggingLevel.INFO

    def _get_message(self) -> str:
        """Gets a message representing the details of the event.

        Returns:
            str: The message for the event.
        """

        return f"Performing action: {self.data['action']!r}\nOn Change: {self.data['change']!r}"



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The Event class represents an event happening in AutoTransform and
serves as a base that can be used to provide consistent typing and
messaging for similar events.
"""

import time
from abc import ABC, abstractmethod
from typing import Any, Generic, Mapping, Optional, TypeVar

from autotransform.event.logginglevel import LoggingLevel
from autotransform.event.type import EventType

TData = TypeVar("TData", bound=Mapping[str, Any])


class Event(Generic[TData], ABC):
    """The base for Events. Used to construct a loggable event that can be hooked in to
    through the EventHandler to store logs in custom deployments.

    Attributes:
        data (TData): The data that represents details of the Event.
        create_time (float): The current timestamp when the event is created.
    """

    data: TData
    create_time: float

    def __init__(self, data: TData):
        """A simple constructor.

        Args:
            data (TData): The data that represents details of the Event.
        """

        self.data = data
        self.create_time = time.time()

    @staticmethod
    @abstractmethod
    def get_type() -> EventType:
        """Used to represent the type of Event, output to logs.

        Returns:
            EventType: The unique type associated with this Event.
        """

    @staticmethod
    @abstractmethod
    def get_logging_level() -> LoggingLevel:
        """The logging level for events of this type.

        Returns:
            LoggingLevel: The logging detail required to log this event.
        """

    @abstractmethod
    def _get_message(self) -> str:
        """Gets a message representing the details of the event.

        Returns:
            str: The message for the event.
        """

    @staticmethod
    def get_color_override() -> Optional[str]:
        """Used to override logging color for specific events where needed. Should use
        colorama ANSI codes.

        Returns:
            Optional[str]: An optional color to use to override defaults when logging.
        """

        return None

    def get_message(self) -> str:
        """Gets a message that can be output to logs representing the event. Converts type from
        lowercase with underscores to capitalized words for the message.

        Returns:
            str: The loggable message.
        """

        type_str = "".join([w.capitalize() for w in self.get_type().split("_")])
        message = self._get_message()
        if message.startswith("["):
            return f"[{type_str}]{message}"
        return f"[{type_str}] {message}"



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the ChangeStateCondition."""

from __future__ import annotations

from typing import ClassVar, List

from autotransform.change.base import Change, ChangeState, ReviewState, TestState
from autotransform.step.condition.base import ComparisonCondition, ConditionName
from autotransform.step.condition.comparison import ComparisonType


class ChangeStateCondition(ComparisonCondition[ChangeState]):
    """A condition which checks the ChangeState against the state supplied using the supplied
    comparison.

    Attributes:
        comparison (ComparisonType): The type of comparison to perform.
        value (ChangeState | List[ChangeState]): The state(s) to compare against.
        name (ClassVar[ConditionName]): The name of the Component.
    """

    comparison: ComparisonType
    value: ChangeState | List[ChangeState]

    name: ClassVar[ConditionName] = ConditionName.CHANGE_STATE

    def get_val_from_change(self, change: Change) -> ChangeState:
        """Gets the state from the Change.

        Args:
            change (Change): The Change the Condition is checking.

        Returns:
            ChangeState: The state of the Change.
        """

        return change.get_state()


class MergeableStateCondition(ComparisonCondition[str]):
    """A condition which checks the mergeable state against the state supplied using the supplied
    comparison.

    Attributes:
        comparison (ComparisonType): The type of comparison to perform.
        value (str | List[str]): The state(s) to compare against.
        name (ClassVar[ConditionName]): The name of the Component.
    """

    comparison: ComparisonType
    value: str | List[str]

    name: ClassVar[ConditionName] = ConditionName.MERGEABLE_STATE

    def get_val_from_change(self, change: Change) -> str:
        """Gets the mergeable state from the Change.

        Args:
            change (Change): The Change the Condition is checking.

        Returns:
            str: The mergeable state of the Change.
        """

        return change.get_mergeable_state()


class ReviewStateCondition(ComparisonCondition[ReviewState]):
    """A condition which checks the ReviewState against the state supplied using the supplied
    comparison.

    Attributes:
        comparison (ComparisonType): The type of comparison to perform.
        value (ReviewState | List[ReviewState]): The state(s) to compare against.
        name (ClassVar[ConditionName]): The name of the Component.
    """

    comparison: ComparisonType
    value: ReviewState | List[ReviewState]

    name: ClassVar[ConditionName] = ConditionName.REVIEW_STATE

    def get_val_from_change(self, change: Change) -> ReviewState:
        """Gets the review state from the Change.

        Args:
            change (Change): The Change the Condition is checking.

        Returns:
            ReviewState: The review state of the Change.
        """

        return change.get_review_state()


class TestStateCondition(ComparisonCondition[TestState]):
    """A condition which checks the TestState against the state supplied using the supplied
    comparison.

    Attributes:
        comparison (ComparisonType): The type of comparison to perform.
        value (TestState | List[TestState]): The state(s) to compare against.
        name (ClassVar[ConditionName]): The name of the Component.
    """

    comparison: ComparisonType
    value: TestState | List[TestState]

    name: ClassVar[ConditionName] = ConditionName.TEST_STATE

    def get_val_from_change(self, change: Change) -> TestState:
        """Gets the state from the Change.

        Args:
            change (Change): The Change the Condition is checking.

        Returns:
            TestState: The test state of the Change.
        """

        return change.get_test_state()



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""An enum representing a type of comparison. Used by Conditions to make determinations
on the type of check to perform. Includes a compare function those Conditions will use."""

from enum import Enum
from typing import Any


class ComparisonType(str, Enum):
    """A list of possible comparisons."""

    # Base Comparisons
    EQUAL = "equal"
    NOT_EQUAL = "not_equal"
    IN = "in"
    NOT_IN = "not_in"

    # Sortable Comparisons
    GREATER_THAN = "greater_than"
    GREATER_THAN_OR_EQUAL = "greater_than_or_equal"
    LESS_THAN = "less_than"
    LESS_THAN_OR_EQUAL = "less_than_or_equal"

    # List Comparisons
    CONTAINS = "contains"
    NOT_CONTAINS = "not_contains"
    EMPTY = "empty"
    NOT_EMPTY = "not_empty"


# pylint: disable=too-many-return-statements
def compare(first_val: Any, second_val: Any, comparison: ComparisonType) -> bool:
    """Performs the comparison specified by a ComparisonType.

    Args:
        first_val (Any): The first value of the comparison.
        second_val (Any): The second value of the comparison.
        comparison (ComparisonType): The type of comparison to perform.

    Raises:
        ValueError: Raised when the comparison is not a valid type.

    Returns:
        bool: The results of the comparison.
    """

    if comparison == ComparisonType.EQUAL:
        return first_val == second_val
    if comparison == ComparisonType.NOT_EQUAL:
        return first_val != second_val
    if comparison == ComparisonType.IN:
        return first_val in second_val
    if comparison == ComparisonType.NOT_IN:
        return first_val not in second_val

    if comparison == ComparisonType.GREATER_THAN:
        return first_val > second_val
    if comparison == ComparisonType.GREATER_THAN_OR_EQUAL:
        return first_val >= second_val
    if comparison == ComparisonType.LESS_THAN:
        return first_val < second_val
    if comparison == ComparisonType.LESS_THAN_OR_EQUAL:
        return first_val <= second_val

    if comparison == ComparisonType.CONTAINS:
        return second_val in first_val
    if comparison == ComparisonType.NOT_CONTAINS:
        return second_val not in first_val
    if comparison == ComparisonType.EMPTY:
        return len(first_val) == 0
    if comparison == ComparisonType.NOT_EMPTY:
        return len(first_val) > 0

    raise ValueError(f"{comparison} is not a valid ComparisonType")



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The run command is used to execute a full run of a schema, either locally
or by kicking off a remote job."""

import json
import os
from argparse import ArgumentParser, Namespace

from autotransform.config import get_config
from autotransform.event.debug import DebugEvent
from autotransform.event.handler import EventHandler
from autotransform.event.logginglevel import LoggingLevel
from autotransform.event.run import ScriptRunEvent
from autotransform.event.verbose import VerboseEvent
from autotransform.filter.base import FACTORY as filter_factory
from autotransform.runner.base import Runner
from autotransform.runner.local import LocalRunner
from autotransform.schema.builder import FACTORY as schema_builder_factory
from autotransform.schema.schema import AutoTransformSchema
from autotransform.util.schema_map import SchemaMap


def add_args(parser: ArgumentParser) -> None:
    """Adds the args to a subparser that are required to run a schema.

    Args:
        parser (ArgumentParser): The parser for the schema run.
    """

    parser.add_argument(
        "schema",
        metavar="schema",
        type=str,
        help="The schema that will be run. Could be a file path, string, "
        + "environment variable name, or builder name.",
    )

    parser.add_argument(
        "--filter",
        metavar="filter",
        type=str,
        help="A JSON encoded filter to add to the schema.",
    )

    parser.add_argument(
        "--max-submissions",
        metavar="max_submissions",
        type=int,
        help="An override to the maximum submissions a schema can produce.",
    )

    logging_level = parser.add_mutually_exclusive_group()
    logging_level.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        required=False,
        help="Tells the script to output verbose logs.",
    )
    logging_level.add_argument(
        "-d",
        "--debug",
        dest="debug",
        action="store_true",
        required=False,
        help="Tells the script to output debug logs.",
    )

    type_group = parser.add_mutually_exclusive_group()
    type_group.add_argument(
        "-b",
        "--builder",
        dest="schema_type",
        action="store_const",
        const="builder",
        required=False,
        help="Tells the script to interpret the schema as a builder name.",
    )
    type_group.add_argument(
        "-f",
        "--file",
        dest="schema_type",
        action="store_const",
        const="file",
        required=False,
        help="Tells the script to interpret the schema as a file path.",
    )
    type_group.add_argument(
        "-s",
        "--string",
        dest="schema_type",
        action="store_const",
        const="string",
        required=False,
        help="Tells the script to interpret the schema as a JSON encoded string",
    )
    type_group.add_argument(
        "-e",
        "--environment",
        dest="schema_type",
        action="store_const",
        const="environment",
        required=False,
        help="Tells the script to interpret the schema as an environment variable storing the JSON "
        + "encoded schema.",
    )
    type_group.add_argument(
        "-n",
        "--name",
        dest="schema_type",
        action="store_const",
        const="name",
        required=False,
        help="Tells the script to interpret the schema as a name stored in schema_map.json.",
    )

    # Run Mode
    mode_group = parser.add_mutually_exclusive_group()
    mode_group.add_argument(
        "-l",
        "--local",
        dest="run_local",
        action="store_true",
        required=False,
        help="Tells the script to run locally, local is the default mode.",
    )
    mode_group.add_argument(
        "-r",
        "--remote",
        dest="run_local",
        action="store_false",
        required=False,
        help="Tells the script to run remote using the remote component from the config.",
    )

    parser.set_defaults(schema_type="file", run_local=True, func=run_command_main)


# pylint: disable=too-many-branches
def run_command_main(args: Namespace) -> None:
    """The main method for the run command, handles the actual execution of a run.

    Args:
        args (Namespace): The arguments supplied to the run command, such as the schema.
    """

    # pylint: disable=unspecified-encoding

    event_handler = EventHandler.get()
    if args.verbose:
        event_handler.set_logging_level(LoggingLevel.VERBOSE)
    if args.debug:
        event_handler.set_logging_level(LoggingLevel.DEBUG)
    schema = args.schema
    event_handler.handle(DebugEvent({"message": f"Schema: ({args.schema_type}) {args.schema}"}))
    event_args = {"schema": args.schema, "schema_type": args.schema_type}
    if args.schema_type == "builder":
        try:
            schema = json.loads(schema)
        except json.JSONDecodeError:
            schema = {"name": schema}
        schema = schema_builder_factory.get_instance(schema).build()
    elif args.schema_type == "file":
        with open(schema, "r") as schema_file:
            schema = AutoTransformSchema.from_data(json.loads(schema_file.read()))
    elif args.schema_type == "environment":
        schema = os.getenv(schema)
        assert isinstance(schema, str)
        schema = AutoTransformSchema.from_data(json.loads(schema))
    elif args.schema_type == "name":
        schema = SchemaMap.get().get_schema(args.schema)
        assert args.schema == schema.config.schema_name
    else:
        schema = AutoTransformSchema.from_data(json.loads(schema))

    repo_override = get_config().repo_override
    if repo_override is not None:
        schema.repo = repo_override

    if args.filter:
        schema.filters.append(filter_factory.get_instance(json.loads(args.filter)))

    if args.max_submissions:
        schema.config.max_submissions = args.max_submissions

    event_handler.handle(VerboseEvent({"message": f"Decoded Schema: {schema!r}"}))

    if args.run_local:
        event_handler.handle(VerboseEvent({"message": "Running locally"}))
        event_args["remote"] = False
        config_runner = get_config().local_runner
        if config_runner is None:
            event_handler.handle(DebugEvent({"message": "No runner defined, using default"}))
            runner: Runner = LocalRunner()
        else:
            runner = config_runner
    else:
        event_handler.handle(VerboseEvent({"message": "Running remote"}))
        event_args["remote"] = True
        config_runner = get_config().remote_runner
        assert config_runner is not None
        runner = config_runner

    event_args["runner"] = runner
    event_handler.handle(ScriptRunEvent({"script": "run", "args": event_args}))
    runner.run(schema)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""AutoTransform is a framework for defining and running code transformations.
It can be used by organizations looking to update/change/cleanup their codebases
as well as by library authors to provide transformations to end users that can
be used to update code for changes to the library (i.e. moving to new APIs).

For more information, see the README.
"""



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the DirectoryBatcher."""

from __future__ import annotations

import pathlib
from copy import deepcopy
from typing import Any, ClassVar, Dict, List, Optional, Sequence

from autotransform.batcher.base import Batch, Batcher, BatcherName
from autotransform.item.base import Item
from autotransform.item.file import FileItem


class DirectoryBatcher(Batcher):
    """A batcher which separates FileItems by directory. Batches are given a title that is based
    on a prefix and the directory that the Batch represents.

    Attributes:
        prefix (str): The prefix to apply to any titles for Batches.
        metadata (Optional[Dict[str, Any]], optional): The metadata to associate with
            Batches. Defaults to None.
        name (ClassVar[BatcherName]): The name of the Component.
    """

    prefix: str
    metadata: Optional[Dict[str, Any]] = None

    name: ClassVar[BatcherName] = BatcherName.DIRECTORY

    def batch(self, items: Sequence[Item]) -> List[Batch]:
        """Takes in a list of FileItems and separates them based on their directory.

        Args:
            items (Sequence[Item]): The filtered Items to separate.

        Returns:
            List[Batch]: A list containing a Batch for each directory containing files.
        """

        # Get a mapping from directory to items within that directory
        item_map: Dict[str, List[FileItem]] = {}
        for item in items:
            assert isinstance(item, FileItem)
            directory = str(pathlib.Path(item.get_path()).parent).replace("\\", "/")
            if directory not in item_map:
                item_map[directory] = []
            item_map[directory].append(item)

        # Create Batches
        batches: List[Batch] = [
            {"items": batch_items, "title": self.prefix + ": " + directory}
            for directory, batch_items in item_map.items()
        ]

        if self.metadata is not None:
            for batch in batches:
                # Deepcopy metadata to ensure mutations don't apply to all Batches
                batch["metadata"] = deepcopy(self.metadata)
        return batches



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""A set of configuration options for a schema."""

from __future__ import annotations

from typing import List, Optional, Type

from pydantic import Field, validator

from autotransform.util.component import ComponentModel
from autotransform.util.console import (
    choose_option,
    choose_options_from_list,
    choose_yes_or_no,
    get_str,
    input_int,
)
from autotransform.validator.base import ValidationResultLevel


class SchemaConfig(ComponentModel):  # pylint: disable=too-few-public-methods
    """An object containing all configuration information for a Schema.

    Attributes:
        schema_name (str): The unique name of the schema.
        allowed_validation_level (ValidationResultLevel, optional): The allowed level of
            validation issues. Any issues raised above this level will trigger exceptions.
            Defaults to ValidationResultLevel.NONE.
        max_submissions (Optional[int], optional): The maximum number of submissions the schema can
            create per run. If None, there is no limit. Defaults to None.
        owners (List[str], optional): The owners for the schema. Defaults to [].
    """

    schema_name: str
    allowed_validation_level: ValidationResultLevel = ValidationResultLevel.NONE
    max_submissions: Optional[int] = None
    owners: List[str] = Field(default_factory=list)

    # pylint: disable=invalid-name
    @validator("max_submissions")
    @classmethod
    def max_submissions_is_positive(cls: Type[SchemaConfig], v: Optional[int]) -> Optional[int]:
        """Validates that max submissions is positive.

        Args:
            cls (Type[SchemaConfig]): The Config class.
            v (int): The maximum number of submissions.

        Raises:
            ValueError: Raised if the maximum number of submissions is not positive.

        Returns:
            Optional[int]: The unmodified maximum number of submissions.
        """

        if v is not None and v < 1:
            raise ValueError(f"Maximum number of submissions must be positive, {v} provided")
        return v

    @staticmethod
    def from_console(prev_config: Optional[SchemaConfig] = None) -> SchemaConfig:
        """Gets a SchemaConfig using console inputs.

        Args:
            prev_config (Optional[SchemaConfig], optional): A previously input SchemaConfig.
                Defaults to None.

        Returns:
            SchemaConfig: The input SchemaConfig.
        """

        if prev_config is not None and choose_yes_or_no(
            f"Use previous schema name ({prev_config.schema_name})?"
        ):
            schema_name = prev_config.schema_name
        else:
            schema_name = get_str("Enter the name of the schema: ")

        if prev_config is not None and choose_yes_or_no(
            f"Use previous allowed validation level ({prev_config.allowed_validation_level.value})?"
        ):
            allowed_validation_level = prev_config.allowed_validation_level
        elif choose_yes_or_no(
            f"Use default allowed validation level ({ValidationResultLevel.NONE})?"
        ):
            allowed_validation_level = ValidationResultLevel.NONE
        else:
            allowed_validation_level = choose_option(
                "Enter allowed validation level",
                [(level, [level.value.lower()]) for level in ValidationResultLevel],
            )

        if prev_config is not None and prev_config.owners:
            owners = choose_options_from_list(
                "Choose owners to keep",
                [(filt, f"{filt!r}") for filt in prev_config.owners],
                min_choices=0,
                max_choices=len(prev_config.owners),
            )
        else:
            owners = []
        extra_owners = get_str("Enter any owners as a comma separated list(blank for none): ")
        owners.extend(owner.strip() for owner in extra_owners.split(","))

        if prev_config is not None and choose_yes_or_no(
            f"Use previous maximum submissions ({prev_config.max_submissions})?"
        ):
            max_submissions = prev_config.max_submissions
        elif choose_yes_or_no("Would you like to limit the maximum number of submissions?"):
            max_submissions = input_int("Enter the maximum number of submissions", min_val=1)
        else:
            max_submissions = None

        return SchemaConfig(
            schema_name=schema_name,
            allowed_validation_level=allowed_validation_level,
            owners=list(set(owners)),
            max_submissions=max_submissions,
        )



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The base class and associated classes for Step components."""

from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import ClassVar, List

from autotransform.change.base import Change
from autotransform.step.action.base import Action
from autotransform.util.component import ComponentFactory, ComponentImport, NamedComponent


class StepName(str, Enum):
    """A simple enum for mapping."""

    CONDITIONAL = "conditional"


class Step(NamedComponent):
    """The base for Step components. Used by AutoTransform to manage outstanding
    Changes, determining what actions to take.

    Attributes:
        name (ClassVar[StepName]): The name of the component.
    """

    name: ClassVar[StepName]

    @abstractmethod
    def get_actions(self, change: Change) -> List[Action]:
        """Checks the Change to determine what actions should be taken. If no Actions
        are returned, the Step is considered skipped.

        Args:
            change (Change): The Change the Step is running against.

        Returns:
            List[Action]: The Actions the Step wants to take.
        """

    @abstractmethod
    def continue_management(self, change: Change) -> bool:
        """Checks if management should be continued after this Step when Actions were provided.

        Args:
            change (Change): The Change the Step is running against.

        Returns:
            bool: Whether to continue management.
        """


FACTORY = ComponentFactory(
    {
        StepName.CONDITIONAL: ComponentImport(
            class_name="ConditionalStep", module="autotransform.step.conditional"
        ),
    },
    Step,  # type: ignore [type-abstract]
    "step.json",
)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""A Change represents a submission from a run of AutoTransform on a particular Batch. They
are used for managing submissions to code review/source control systems. A Pull Request is
an example of a potential Change."""



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Enums used by utilities of AutoTransform or across a number of components."""

from enum import Enum


class AggregatorType(str, Enum):
    """A list of possible comparisons."""

    ALL = "all"
    ANY = "any"


class SchemaType(str, Enum):
    """Possible types of Schemas to use."""

    BUILDER = "builder"
    FILE = "file"
